VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cVolume"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Attribute VB_Ext_KEY = "SavedWithClassBuilder6" ,"Yes"
Attribute VB_Ext_KEY = "Top_Level" ,"Yes"
Option Explicit

Public Enum RECORDS_TREE_FILTER
   rtfUndefined = -1
   rtfAllRecords = 0
   rtfDeleted = 1
   rtfOrphans = 2
End Enum

Private Enum RECORD_TYPES
   rtEmpty = 0
   rtInvalid = 1
   rtResident = 2
   rtMetaFile = 4
   rtRecordOK = 8
   rtOrphan = 16
End Enum

Private Type CLUSTERS_BLOCK
   blockIndex As Long
   startLCN As Long
   endLCN As Long
   attrType As Byte
   mftRecordNum As Long
   crossLinksCount As Long
   crossLinkIndex() As Long
End Type

Private Type DELETED_PARENT
   Children As Long
   Child() As Long
End Type

Public memChunkSize As Long
Public recChunkSize As Long
Public cbChunkSize As Long

Private Const default_memChunkSize As Long = 1024# * 1024 ' 1MB chunks
Private Const default_recChunkSize As Long = 32# * 1024   ' read 32K records at time
Private Const default_cbChunkSize As Long = 128# * 1024   ' Initial size of aClusters array
Private Const default_delParents As Long = 32# * 1024    ' Initial size of aClusters array

Public Event Progress(ByVal percent As Long)
Public Event StatusChanged(ByVal strStatus As String)

Private m_Info As cVolumeInfo
Private m_Filter As RECORDS_TREE_FILTER

Private bInit As Boolean, bReadMFT As Boolean, bScanning As Boolean
Private clusterBlocksCount As Long, deletedParentsCount As Long
Private hDevice As Long
Private m_DriveLetter As String
Private m_fMapClusters As Boolean
Private aRecords() As MFT_RECORD_INFO
Private aClusters() As CLUSTERS_BLOCK
Private aBitmap() As Byte
Private abMFTBitmap() As Byte
Private aDeleted() As DELETED_PARENT

Public Property Get info() As cVolumeInfo
   Set info = m_Info
End Property

Public Property Let MapClusters(ByVal bMap As Boolean)
   m_fMapClusters = bMap
End Property

Public Property Get MapClusters() As Boolean
   MapClusters = m_fMapClusters
End Property

Public Property Let DriveLetter(ByVal strLetter As String)
   If strLetter = "" Then Exit Property
   If m_DriveLetter = UCase(Left(strLetter, 1)) Then Exit Property
   bScanning = True
   m_DriveLetter = UCase(Left(strLetter, 1))
   bReadMFT = False
   bInit = False
   Erase aRecords
   Erase aClusters
   Erase aBitmap
   hDevice = OpenVolume
   If hDevice = INVALID_HANDLE_VALUE Then
      MsgBox "Error initializing volume " & m_DriveLetter & vbCrLf & ApiError(GetLastError), vbCritical, "Volume access error"
      m_DriveLetter = ""
      Exit Property
   End If
   Dim ret As Long
   ret = m_Info.InitVolumeInfo(hDevice)
   If ret <> 0 Then
      MsgBox "Error getting volume info" & m_DriveLetter & vbCrLf & ApiError(ret), vbCritical, "Volume access error"
      m_DriveLetter = ""
      Exit Property
   End If
'*****Initialize arrays and variables*****
   ReDim aRecords(m_Info.MFTRecordsCount - 1)
   ReDim aClusters(cbChunkSize - 1)
   ret = InitVolumeBitMap
   If ret <> 0 Then
      MsgBox "Error reading volume bitmap" & m_DriveLetter & vbCrLf & ApiError(ret), vbCritical, "Volume access error"
      m_DriveLetter = ""
      Exit Property
   End If
   bScanning = False
   bInit = True
End Property

Public Property Get DriveLetter() As String
   DriveLetter = m_DriveLetter
End Property

Public Property Get MFTReadOK() As Boolean
   MFTReadOK = bReadMFT
End Property

Public Property Let RecordsTreeFilter(ByVal filter As RECORDS_TREE_FILTER)
    m_Filter = filter
End Property

Public Property Get RecordsTreeFilter() As RECORDS_TREE_FILTER
   RecordsTreeFilter = m_Filter
End Property

Public Function OpenVolume() As Long
   OpenVolume = CreateFile("\\.\" & m_DriveLetter & ":", GENERIC_READ Or GENERIC_WRITE, FILE_SHARE_READ Or FILE_SHARE_WRITE, 0&, OPEN_EXISTING, 0&, 0&)
End Function

'*********************Direct Read/Write staff*****************************************

'===Direct read volume data===
'iStartSec     - start logical sector to read (IN)
'iOffset       - offset from start sector in bytes (IN)
'cBytes        - number of bytes to read (IN)
'abData        - bytes array to be filled with data (OUT)
'bShowProgress - boolean flag to raise (or not) progress events (IN).
Public Function DirectRead(ByVal iStartSector As Variant, ByVal iOffset As Long, _
                           ByVal cBytes As Long, ByRef abData() As Byte, _
                           Optional bShowProgress As Boolean = False) As Long
'*****Sanitize*****
    If Not bInit Then
       DirectRead = 21 'Device not ready error
       Exit Function
    End If
    If (cBytes <= 0) Or (iStartSector < 0) Then
       DirectRead = 87 'Invalid parameter error
       Exit Function
    End If

'*****Normalize offset*****
    Dim nSectors As Long
    If iOffset >= m_Info.BytesPerSector Then
       nSectors = Int(iOffset / m_Info.BytesPerSector)
       iStartSector = iStartSector + nSectors
       iOffset = iOffset - nSectors * m_Info.BytesPerSector
    End If

'*****Setting file pointer to required offset*****
    Dim vOffset As Variant
    Dim dwLow As Long, dwHigh As Long
    vOffset = CDec(CDec(iStartSector) * m_Info.BytesPerSector)
    dwLow = LoDWord(vOffset)
    dwHigh = HiDWord(vOffset)
    If SetFilePointer(hDevice, dwLow, dwHigh, FILE_BEGIN) = INVALID_HANDLE_VALUE Then
       DirectRead = GetLastError
       Exit Function
    End If
    
'*****Reading data by chunks*****
    Dim abBuff() As Byte, abTemp() As Byte
    Dim nRead As Long, ret As Long
    Dim bytesRemain As Long, i As Long, nChunks As Long
    
    nSectors = Int((iOffset + cBytes - 1) / m_Info.BytesPerSector) + 1
    bytesRemain = nSectors * m_Info.BytesPerSector
    nChunks = Int(cBytes / memChunkSize)
    ReDim abBuff(memChunkSize - 1)
    ReDim abTemp(bytesRemain - 1)
    i = 0
    Do While bytesRemain > memChunkSize
       If ReadFile(hDevice, abBuff(0), memChunkSize, nRead, 0&) = 0 Then
          DirectRead = GetLastError
          GoTo Free_Arrays
       End If
       CopyMemory abTemp(i * memChunkSize), abBuff(0), memChunkSize
       If bShowProgress Then RaiseEvent Progress(i / nChunks * 100)
       bytesRemain = bytesRemain - memChunkSize
       i = i + 1
    Loop
    If bytesRemain > 0 Then
        ReDim abBuff(bytesRemain - 1)
        If ReadFile(hDevice, abBuff(0), bytesRemain, nRead, 0&) = 0 Then
          DirectRead = GetLastError
          GoTo Free_Arrays
        End If
        CopyMemory abTemp(i * memChunkSize), abBuff(0), bytesRemain
    End If
    Erase abBuff
    ReDim abData(cBytes - 1)
    CopyMemory abData(0), abTemp(iOffset), cBytes
    DirectRead = 0
Free_Arrays:
    Erase abBuff
    Erase abTemp
    If bShowProgress Then
       RaiseEvent Progress(100)
       RaiseEvent StatusChanged("Done")
    End If
End Function

'===Direct write data===
'iStartSec     - start logical sector to write (IN)
'iOffset       - offset from start sector (IN)
'sWrite        - string to be written (IN)
'bShowProgress - boolean flag to raise (or not) progress events (IN,Opional,Default=False).

Public Function DirectWrite(ByVal iStartSector As Variant, ByVal iOffset As Long, ByVal sWrite As String, Optional bShowProgress As Boolean = False) As Long
'*****Sanitize*****
    If Not bInit Then
       DirectWrite = 21 'Device not ready error
       Exit Function
    End If
    If (Len(sWrite) = 0) Or (iStartSector < 0) Then
       DirectWrite = 87 'Invalid parameter error
       Exit Function
    End If

'*****Normalize offset*****
    Dim nSectors As Long
    If iOffset >= m_Info.BytesPerSector Then
       nSectors = Int(iOffset / m_Info.BytesPerSector)
       iStartSector = iStartSector + nSectors
       iOffset = iOffset - nSectors * m_Info.BytesPerSector
    End If
    
'*****Setting file pointer to required offset*****
    Dim vOffset As Variant, vSize As Variant
    Dim offsetLow As Long, offsetHigh As Long
    vOffset = CDec(CDec(iStartSector) * m_Info.BytesPerSector)
    offsetLow = LoDWord(vOffset)
    offsetHigh = HiDWord(vOffset)
    If SetFilePointer(hDevice, offsetLow, offsetHigh, FILE_BEGIN) = INVALID_HANDLE_VALUE Then
       DirectWrite = GetLastError
       Exit Function
    End If
    
    Dim abBuff() As Byte
    Dim abTemp() As Byte
    
'*****Reading first sector*****
    Dim ret As Long, nWrite As Long
    ReDim abBuff(m_Info.BytesPerSector - 1)
    ret = ReadFile(hDevice, abBuff(0), m_Info.BytesPerSector, nWrite, 0&)
    If ret <> 0 Then
       DirectWrite = ret
       Erase abBuff
       Exit Function
    End If
    'Reset file pointer after reading
    If SetFilePointer(hDevice, offsetLow, offsetHigh, FILE_BEGIN) = INVALID_HANDLE_VALUE Then
       DirectWrite = GetLastError
       Exit Function
    End If
    
    abTemp = StrConv(Left(sWrite, m_Info.BytesPerSector - iOffset), vbFromUnicode)
    CopyMemory abBuff(iOffset), abTemp(0), UBound(abTemp) + 1
    Erase abTemp
    
    Dim sizeLow As Long, sizeHigh As Long
    nSectors = Int((iOffset + Len(sWrite) - 1) / m_Info.BytesPerSector) + 1
    vSize = CDec(CDec(nSectors) * m_Info.BytesPerSector)
    sizeLow = LoDWord(vSize)
    sizeHigh = HiDWord(vSize)
    
    Call LockFile(hDevice, offsetLow, offsetHigh, sizeLow, sizeHigh)
'*****Write first sector back*****
    If WriteFile(hDevice, abBuff(0), UBound(abBuff) + 1, nWrite, 0&) = 0 Then
       DirectWrite = GetLastError
       GoTo Unlock_Device
    End If
'*****Write remain data by chunks*****
    Dim i As Long, nChunks As Long
    Dim bytesRemain As Variant
    bytesRemain = CDec(CDec(nSectors) * m_Info.BytesPerSector) - m_Info.BytesPerSector
    nChunks = Int(Len(sWrite) / memChunkSize)
    
    Do While bytesRemain > memChunkSize
       abBuff = StrConv(Mid(sWrite, i * memChunkSize + 1, memChunkSize), vbFromUnicode)
       If WriteFile(hDevice, abBuff(0), UBound(abBuff) + 1, nWrite, 0&) = 0 Then
          DirectWrite = GetLastError
          GoTo Unlock_Device
       End If
       bytesRemain = bytesRemain - memChunkSize
       i = i + 1
       If bShowProgress Then RaiseEvent Progress(i / nChunks * 100)
    Loop
'*****Write remain data (sectors after last chunk)*****
    If bytesRemain > 0 Then
       abBuff = StrConv(Mid(sWrite, i * memChunkSize + 1, bytesRemain), vbFromUnicode)
       If WriteFile(hDevice, abBuff(0), UBound(abBuff) + 1, nWrite, 0&) = 0 Then
          DirectWrite = GetLastError
          GoTo Unlock_Device
       End If
    End If
    Call FlushFileBuffers(hDevice)
Unlock_Device:
    Call UnlockFile(hDevice, offsetLow, offsetHigh, sizeLow, sizeHigh)
    Erase abBuff
    If bShowProgress Then
       RaiseEvent Progress(100)
       RaiseEvent StatusChanged("Done")
    End If
End Function

'===Direct fill clusters with single byte===
'startLCN      - start logical cluster to fill (IN)
'nClusters     - number of clusters to be filled (IN)
'byteToWrite   - single byte to fill clusters with (IN)
'bShowProgress - boolean flag to raise (or not) progress events (IN,Opional,Default=False).
Public Function DirectFill(ByVal startLCN As Long, ByVal nClusters As Long, ByVal byteToWrite As Byte, Optional bShowProgress As Boolean = False) As Long
'Sanitize
    If Not bInit Then
       DirectFill = 21 'Device not ready error
       Exit Function
    End If
    If (nClusters <= 0) Or (startLCN < 0) Then
       DirectFill = 87 'Invalid parameter error
       Exit Function
    End If
    
    Dim abBuff() As Byte
    Dim offsetLow As Long, offsetHigh As Long
    Dim vOffset As Variant, vSize As Variant
    vOffset = CDec(CDec(startLCN) * m_Info.BytesPerCluster)
    offsetLow = LoDWord(vOffset)
    offsetHigh = HiDWord(vOffset)
    If SetFilePointer(hDevice, offsetLow, offsetHigh, FILE_BEGIN) = INVALID_HANDLE_VALUE Then
        DirectFill = GetLastError
        Exit Function
    End If
    Dim sizeLow As Long, sizeHigh As Long
    vSize = CDec(CDec(nClusters) * m_Info.BytesPerCluster)
    sizeLow = LoDWord(vSize)
    sizeHigh = HiDWord(vSize)
    Call LockFile(hDevice, offsetLow, offsetHigh, sizeLow, sizeHigh)
    
    Dim cBytes As Long, nWrite As Long, i As Long, nChunks As Long
    Dim bytesRemain As Variant
    
    cBytes = memChunkSize
    bytesRemain = CDec(CDec(nClusters) * m_Info.BytesPerCluster)
    nChunks = Int(bytesRemain / memChunkSize)
    ReDim abBuff(cBytes - 1)
    FillMemory abBuff(0), cBytes, byteToWrite
    Do While bytesRemain > memChunkSize
       If WriteFile(hDevice, abBuff(0), cBytes, nWrite, 0&) = 0 Then
          DirectFill = GetLastError
          GoTo Unlock_Device
       End If
       bytesRemain = bytesRemain - memChunkSize
       i = i + 1
       If bShowProgress Then RaiseEvent Progress(i / nChunks * 100)
    Loop
    If bytesRemain > 0 Then
       cBytes = bytesRemain
       ReDim Preserve abBuff(cBytes - 1)
       If WriteFile(hDevice, abBuff(0), cBytes, nWrite, 0&) = 0 Then
          DirectFill = GetLastError
          GoTo Unlock_Device
       End If
    End If
    Call FlushFileBuffers(hDevice)
Unlock_Device:
    Call UnlockFile(hDevice, offsetLow, offsetHigh, sizeLow, sizeHigh)
    Erase abBuff
    If bShowProgress Then
       RaiseEvent Progress(100)
       RaiseEvent StatusChanged("Done")
    End If
End Function

'===Direct overwrite clusters with string pattern===
'startLCN      - start logical cluster to fill (IN)
'nClusters     - number of clusters to be filled (IN)
'strPattern    - string pattern to overwrite clusters with (IN)
'bShowProgress - boolean flag to raise (or not) progress events (IN,Opional,Default=False).
Public Function DirectOverWrite(ByVal startLCN As Variant, ByVal nClusters As Variant, ByVal strPattern As String, Optional bShowProgress As Boolean = False) As Long
    If Not bInit Then
       DirectOverWrite = 21 'Device not ready error
       Exit Function
    End If
    If (nClusters <= 0) Or (startLCN < 0) Or (Len(strPattern) = 0) Then
       DirectOverWrite = 87 'Invalid parameter error
       Exit Function
    End If
    Dim abBuff() As Byte
    Dim offsetLow As Long, offsetHigh As Long
    Dim vOffset As Variant, vSize As Variant
    vOffset = CDec(CDec(startLCN) * m_Info.BytesPerCluster)
    offsetLow = LoDWord(vOffset)
    offsetHigh = HiDWord(vOffset)
    If SetFilePointer(hDevice, offsetLow, offsetHigh, FILE_BEGIN) = INVALID_HANDLE_VALUE Then
        DirectOverWrite = GetLastError
        Exit Function
    End If
    
    Dim sizeLow As Long, sizeHigh As Long
    vSize = CDec(CDec(nClusters) * m_Info.BytesPerCluster)
    sizeLow = LoDWord(vSize)
    sizeHigh = HiDWord(vSize)
    Call LockFile(hDevice, offsetLow, offsetHigh, sizeLow, sizeHigh)
    
    Dim cBytes As Long, nChunks As Long, nWrite As Long, i As Long
    Dim bytesRemain As Variant
    
    cBytes = memChunkSize
    bytesRemain = CDec(CDec(nClusters) * m_Info.BytesPerCluster)
    nChunks = Int(bytesRemain / memChunkSize)
    abBuff = MakeBytesPattern(cBytes, strPattern)
    Do While bytesRemain > memChunkSize
       If WriteFile(hDevice, abBuff(0), cBytes, nWrite, 0&) = 0 Then
          DirectOverWrite = GetLastError
          GoTo Unlock_Device
       End If
       bytesRemain = bytesRemain - memChunkSize
       i = i + 1
       If bShowProgress Then RaiseEvent Progress(i / nChunks * 100)
    Loop
    If bytesRemain > 0 Then
       cBytes = bytesRemain
       ReDim Preserve abBuff(cBytes - 1)
       If WriteFile(hDevice, abBuff(0), cBytes, nWrite, 0&) = 0 Then
          DirectOverWrite = GetLastError
          GoTo Unlock_Device
       End If
    End If
    Call FlushFileBuffers(hDevice)
Unlock_Device:
    Call UnlockFile(hDevice, offsetLow, offsetHigh, sizeLow, sizeHigh)
    Erase abBuff
    If bShowProgress Then
       RaiseEvent Progress(100)
       RaiseEvent StatusChanged("Done")
    End If
End Function

'======Create pattern to fill volume area======
'cBytes - size of area to fill with pattern
'strPattern - initial patternt
Private Function MakeBytesPattern(ByVal cBytes As Long, ByVal strPattern As String) As Byte()
'*****Sanitize*****
    If (cBytes = 0) Or (Len(strPattern) = 0) Then Exit Function
    Dim nPatterns As Long, i As Long
    Dim arrPatterns() As String
    nPatterns = Int(cBytes / Len(strPattern)) + 1
    ReDim arrPaterns(nPatterns - 1)
    For i = 0 To nPatterns - 1
        arrPatterns(i) = strPattern
    Next i
    strPattern = Left(Join(arrPatterns, ""), cBytes)
    MakeBytesPattern = StrConv(strPattern, vbFromUnicode)
End Function

Public Function OverWriteDeleted(Optional ByVal bUsePattern As Boolean = False, _
                                 Optional ByVal byteFill As Byte = 0, _
                                 Optional ByRef colRecords As Collection = Nothing, _
                                 Optional ByVal bSkipCrossLinked As Boolean = True) As Long
   Dim i As Long, ret As Long
   Dim adr As ATTR_DATA_RUNS
   Dim abRecord() As Byte
   If colRecords Is Nothing Then
      For i = 0 To UBound(aRecords)
          With aRecords(i)
              If (.Flags And 1) = 0 Then 'deleted
                 If .Flags And 2 Then 'folder
                    
                 Else 'file
                    ReadMFTRecord i, abRecord
                    
                 End If
              End If
          End With
      Next i
   End If
End Function

'****************MFT records staff*********************
'=====Scan ALL MFT records (by 10K chunks)====
Public Function ScanMFT(Optional bShowProgress As Boolean = False) As Long
   Dim i As Long, ret As Long, nRecord As Long
   Dim recordOffset As Long, recChunks As Long, recordsRemain As Long
   clusterBlocksCount = 0
   recordsRemain = m_Info.MFTRecordsCount
   
   bScanning = True

'*****Scan recChunkSize (64K records) at time*****
   Do While recordsRemain > recChunkSize
      ret = ScanMFTRecords(nRecord, recChunkSize, bShowProgress)
      If ret <> 0 Then
         ScanMFT = ret
         bScanning = False
         Exit Function
      End If
      nRecord = nRecord + recChunkSize
      recordsRemain = recordsRemain - recChunkSize
   Loop
   If recordsRemain > 0 Then
      ret = ScanMFTRecords(nRecord, recordsRemain, bShowProgress)
      If ret <> 0 Then
         ScanMFT = ret
         bScanning = False
         Exit Function
      End If
   End If

'*****Sorting clusters*****
   RaiseEvent StatusChanged("Sorting clusters")
   Call SortClustersBlocks

'*****Cross-links checking*****
   RaiseEvent StatusChanged("Cross-links checking")
   Call CheckCrossLinks
   
   EnumDeletedParents
   
   RaiseEvent StatusChanged("Done")
   
   bReadMFT = True
   bScanning = False
End Function

'=====Scan some MFT records (by 10K chunks)====
'startRecord - first record to scan
'nRecords - # records to scan
Public Function ScanMFTRecords(ByVal startRecord As Long, ByVal nRecords As Long, Optional bShowProgress As Boolean = False) As Long
   Dim ret As Long
   Dim abMFT() As Byte
'*****Read nRecords starting from startRecord*****
   RaiseEvent StatusChanged("Reading MFT (" & startRecord & " - " & startRecord + nRecords & " records of " & Format(m_Info.MFTRecordsCount, "#,0") & " total records)...")
   ret = ReadMFT(startRecord, nRecords, abMFT, bShowProgress)
   If ret <> 0 Then
      ScanMFTRecords = ret
      Exit Function
   End If
'*****FixUp records*****
   RaiseEvent StatusChanged("Fixup MFT(" & startRecord & " - " & startRecord + nRecords & " records of " & Format(m_Info.MFTRecordsCount, "#,0") & " total records)...")
   ret = FixUpMFT(abMFT, 0, nRecords, bShowProgress)
   If ret <> 0 Then
      ScanMFTRecords = ret
      Exit Function
   End If
'*****Parsing records*****
   RaiseEvent StatusChanged("Parsing MFT(" & startRecord & " - " & startRecord + nRecords & " records of " & Format(m_Info.MFTRecordsCount, "#,0") & " total records)...")
   ret = ParseMFT(abMFT, startRecord, nRecords, bShowProgress)
   If ret <> 0 Then
      ScanMFTRecords = ret
      Exit Function
   End If
   ReDim Preserve aClusters(clusterBlocksCount - 1)
   RaiseEvent StatusChanged("Done")
   Erase abMFT
End Function

'=====Read MFT records ====
'startRecord - first record to read
'nRecords - # records to read
'abData() - byte array to be filled with MFT records data
Public Function ReadMFT(ByVal startRecord As Long, _
                        ByVal nRecords As Long, _
                        ByRef abData() As Byte, _
                        Optional bShowProgress As Boolean = False) As Long
    If (nRecords <= 0) Or (startRecord < 0) Then
       ReadMFT = 87 'Invalid parameter error
       Exit Function
    End If
    Dim startSector As Long, cBytes As Long
    With m_Info
       startSector = startRecord * .SectorsPerMFTRecord + .MFTStartLcn * .SectorsPerCluster
       cBytes = nRecords * .BytesPerMFTRecord
    End With
    ReadMFT = DirectRead(startSector, 0, cBytes, abData, bShowProgress)
End Function

'=====Read ONE MFT records ====
'recordNum - record number to read
'abRecord() - byte array to be filled with MFT record data
Public Function ReadMFTRecord(ByVal recordNum As Long, ByRef abRecord() As Byte) As Long
    ReadMFTRecord = ReadMFT(recordNum, 1, abRecord, False)
End Function

'=====Fix up MFT records (update last 2 bytes of earch record's sector)====
'abRecords() - byte array to be updated
'startRecord - first record to FixUp
'nRecords - # records to FixUp
Private Function FixUpMFT(ByRef abRecords() As Byte, _
                          Optional ByVal startRecord As Long = 0, _
                          Optional ByVal nRecords As Long = 0, _
                          Optional bShowProgress As Boolean) As Long
   Dim i As Long, j As Long
   Dim fixupOffset As Integer, fixupSize As Integer, recordOffset As Long
   Dim sectorSize As Long, mftRecordSize As Long
   Dim fixups() As Integer
   
   If startRecord > 0 Then recordOffset = startRecord * m_Info.BytesPerMFTRecord
   If nRecords = 0 Then
      nRecords = (UBound(abRecords) + 1 - recordOffset) / m_Info.BytesPerMFTRecord
   End If
   sectorSize = m_Info.BytesPerSector
   mftRecordSize = m_Info.BytesPerMFTRecord
   For i = 1 To nRecords
      CopyMemory fixupOffset, abRecords(recordOffset + 4), 2
      CopyMemory fixupSize, abRecords(recordOffset + 6), 2
      If fixupSize > m_Info.SectorsPerMFTRecord + 1 Then
         FixUpMFT = 87 'Invalid parameter
         Exit Function
      End If
      If fixupSize > 1 Then
         ReDim fixups(fixupSize - 2)
         CopyMemory fixups(0), abRecords(recordOffset + fixupOffset + 2), (fixupSize - 1) * 2
         For j = 0 To fixupSize - 2
             CopyMemory abRecords(recordOffset + sectorSize + j * sectorSize - 2), fixups(j), 2
         Next j
      End If
      recordOffset = recordOffset + mftRecordSize
   Next i
End Function

'=====Fix up ONE MFT record (update last 2 bytes of earch sector)====
'abRecords() - byte array to be updated
Public Function FixUpMFTRecord(ByRef abRecord() As Byte) As Long
   FixUpMFTRecord = FixUpMFT(abRecord, 0, 1)
End Function

'=====Parse MFT records (fill aRecords array and clusters hash table)====
'abRecords() - byte array to be parsed
'startRecord - first record to parse
'nRecords - # records to parse
Private Function ParseMFT(ByRef abRecords() As Byte, _
                          Optional ByVal startRecord As Long = 0, _
                          Optional ByVal nRecords As Long = 1, _
                          Optional bShowProgress As Boolean = False) As Long
   Dim i As Long, ret As Long
   For i = 0 To nRecords - 1
       ret = ParseMFTRecord(abRecords, aRecords(startRecord + i), i)
       If ret <> 0 Then
          ParseMFT = ret
          Exit For
       End If
       If (i Mod 1000) = 0 Then
          If bShowProgress Then RaiseEvent Progress((i + 1) * 100 / nRecords)
       End If
   Next i
End Function

'=====Parse ONE MFT record (fill aRecords array and clusters hash table)====
'abRecords() - byte array with MFT record(s) to be parsed
'recNum - MFT record to be parsed (relative to abRecords() start)
'record - MFT_RECORD_INFO structure to be filled with data

Private Function ParseMFTRecord(ByRef abRecord() As Byte, ByRef record As MFT_RECORD_INFO, Optional ByVal recNum As Long = 0) As Long
   Dim recordOffset As Long
   recordOffset = recNum * m_Info.BytesPerMFTRecord
   If abRecord(recordOffset) <> &H46 Then
      If abRecord(recordOffset) <> 0 Then record.Type = rtInvalid
      Exit Function
   End If
   Dim hdr As FILE_RECORD_HEADER
   CopyMemory hdr, abRecord(recordOffset), Len(hdr)
   record.nodeRef.dwNumLow = hdr.dwMFTRecordNumber
   record.nodeRef.wSequence = hdr.wSequenceNumber
   record.mainRef = hdr.BaseFileRecord
   record.Flags = hdr.wFlags
   If hdr.wAttributeOffset <= 0 Then
      Exit Function
   End If
   
   Dim attrOffset As Integer
   Dim bNonResident As Boolean
   Dim pos As Long
   Dim dataRunsOffset As Long, dataRunsSize As Long
   
   Dim ah As ATTR_HEADER
   Dim attrResident As ATTR_RESIDENT
   Dim attrNonResident As ATTR_NONRESIDENT
   Dim attrStandard As ATTR_STANDARD
   Dim attrFileName As ATTR_FILENAME
   Dim adr As ATTR_DATA_RUNS
   
   pos = hdr.wAttributeOffset
   Do
      CopyMemory ah, abRecord(recordOffset + pos), Len(ah)
      If ah.dwType = &HFFFFFFFF Then Exit Do
      If ah.uchNonResFlag Then
         bNonResident = True
         CopyMemory attrNonResident, abRecord(recordOffset + pos + Len(ah)), Len(attrNonResident)
         If hdr.dwMFTRecordNumber = 0 Then '$MFT
            'Read $MFT:$BITMAP
            If ah.dwType = &HB0 Then 'MFT bitmap
               adr.attrType = ah.dwType
               adr.allocatedDataSize = LI_Decimal_LE(attrNonResident.n64AllocSize)
               adr.realDataSize = LI_Decimal_LE(attrNonResident.n64RealSize)
               adr.StartVCN = LI_Decimal_LE(attrNonResident.n64StartVCN)
               adr.EndVCN = LI_Decimal_LE(attrNonResident.n64EndVCN)
               Call GetDataRuns(abRecord, recordOffset + pos + attrNonResident.wDatarunOffset, adr, record.nodeRef.dwNumLow)
               ReDim abMFTBitmap(0)
               ExtractDataRuns adr, abMFTBitmap
            End If
         Else
            If m_fMapClusters Then
               'Get data runs for non-resident attribute
               adr.attrType = ah.dwType
               adr.allocatedDataSize = LI_Decimal_LE(attrNonResident.n64AllocSize)
               adr.realDataSize = LI_Decimal_LE(attrNonResident.n64RealSize)
               adr.StartVCN = LI_Decimal_LE(attrNonResident.n64StartVCN)
               adr.EndVCN = LI_Decimal_LE(attrNonResident.n64EndVCN)
               Call GetDataRuns(abRecord, recordOffset + pos + attrNonResident.wDatarunOffset, adr, record.nodeRef.dwNumLow)
               If adr.drCount > 0 Then Call MarkClusters(hdr.dwMFTRecordNumber, adr)
            End If
         End If
      End If
      If ah.dwFullLength = 0 Then Exit Do
      If record.Type = rtEmpty Then record.Type = rtRecordOK
      Select Case ah.dwType
          Case 0      'UNUSED (corrupted?)
          Case &H10   'STANDARD_INFORMATION. Always resident. 48 to 72 bytes
          Case &H20   'ATTR_LIST
          Case &H30   'FILE_NAME. Always resident.
               CopyMemory attrResident, abRecord(recordOffset + pos + Len(ah)), Len(attrResident)
               CopyMemory attrFileName, abRecord(recordOffset + pos + attrResident.wAttrOffset), attrResident.dwLength
               record.FileName = Left(StrConv(attrFileName.wFilename, vbFromUnicode), attrFileName.chFileNameLength)
               record.parentRef = attrFileName.dwMftParentDir
               If record.parentRef.wSequence > record.nodeRef.wSequence Then
                  If (record.Flags And 1) = 0 Then record.Type = record.Type Or rtOrphan
               End If
          Case &H40   'OBJECT_ID. Always resident.
          Case &H50   'SECURITY_DESCRIPTOR
          Case &H60   'VOLUME_NAME. Always resident.
          Case &H70   'VOLUME_INFORMATION. Always resident.
          Case &H80   'DATA
          Case &H90   'INDEX_ROOT. Always resident.
          Case &HA0   'INDEX_ALLOCATION
          Case &HB0   'BITMAP
          Case &HC0   'REPARSE_POINT
          Case &HD0   'EA_INFORMATION
          Case &HE0   'EA_DATA
          Case &HF0   'PROPERTY_SET
          Case &H100  'LOGGED_UTILITY_STREAM
          Case &H1000 'FIRST_USER_DEFINED_ATTRIBUTE
          Case &HFFFF&, &HFFFFFFFF ' END
               Exit Do
      End Select
      pos = pos + ah.dwFullLength
      If pos > m_Info.BytesPerMFTRecord - Len(ah) Then Exit Do
   Loop
   If Not bNonResident Then record.Type = record.Type Or rtResident
End Function

'=====Get data runs for MFT record attribute====
'abRecord() - byte array with MFT record(s) to be parsed
'drOffset - offset to data runs sequence (including recordOffset if abRecord() contains more then 1 record
'adr - ATTR_DATA_RUNS structure to be filled with data runs
'mftRecNum - MFT record number
Private Function GetDataRuns(ByRef abRecord() As Byte, _
                             ByVal drOffset As Long, _
                             ByRef adr As ATTR_DATA_RUNS, _
                             Optional ByVal mftRecNum As Long, _
                             Optional bSkipSparseFiles As Boolean = False) As Long
   Dim drSize As Variant
   Dim LCN As Variant, vcn As Variant, nClusters As Variant
   Dim sizeDataLength As Long, vcnDataLength As Long
   Dim liClusters As LARGE_INTEGER, liVCN As LARGE_INTEGER
   Dim pos As Long, i As Long, j As Long
   Dim b As Byte
   Dim abVCN() As Byte
   Dim bSigned As Boolean, bSparse As Boolean
   
   pos = drOffset
   ReDim adr.dataRuns(0)
   Do
     b = abRecord(pos)
     If b = 0 Then Exit Do
     If b = &HFF Then Exit Do
     sizeDataLength = b And &HF         'First nibble  - bytes allocated for data run lenth value
     vcnDataLength = (b And &HF0) / 16  'Second nibble - bytes allocated for VCN value (0 for sparse data run)
     If sizeDataLength > 8 Then Exit Do 'Exceed 8 bit integer (dr corrupted?)
     If vcnDataLength > 8 Then vcnDataLength = 0 'Exceed 8 bit integer (dr corrupted?). Some sparse records have FF values - skip.
     bSparse = (vcnDataLength = 0)
     ZeroMemory liClusters, 8
     ZeroMemory liVCN, 8
     If i > 0 Then
        ReDim Preserve adr.dataRuns(i)
        ReDim abVCN(7)
        If Not bSparse Then
           CopyMemory abVCN(0), abRecord(pos + sizeDataLength + 1), vcnDataLength
           'vcn are signed variables, so check for negative bit
           If abVCN(vcnDataLength - 1) And &H80 Then 'negative
              For j = 7 To vcnDataLength Step -1
                  abVCN(j) = &HFF
              Next j
              bSigned = True
           Else
              bSigned = False
           End If
        End If
        CopyMemory liVCN, abVCN(0), 8
        vcn = LI_Decimal_LE(liVCN, bSigned)
     Else
        CopyMemory liVCN, abRecord(pos + sizeDataLength + 1), vcnDataLength
        vcn = 0
        'vcn = adr.StartVCN
        LCN = LI_Decimal_LE(liVCN)
     End If
     If sizeDataLength > 0 Then
        CopyMemory liClusters, abRecord(pos + 1), sizeDataLength
     End If
     
     LCN = LCN + vcn
     'Skip invalid LCN's (accept zero LCN for MFT record #7 = $BOOT) only.
     If (LCN <= m_Info.TotalClusters) And ((LCN > 0) Or mftRecNum = 7) Then
        nClusters = LI_Decimal_LE(liClusters)
        If nClusters < 0 Then nClusters = 0
        If nClusters > m_Info.TotalClusters Then nClusters = 0
        adr.dataRuns(i).nClusters = nClusters
        adr.dataRuns(i).LCN = LCN
        adr.dataRuns(i).vcn = vcn
        adr.dataRuns(i).bSparse = bSparse
     End If
     i = i + 1
     pos = pos + sizeDataLength + vcnDataLength + 1
   Loop
   adr.drCount = i
   Erase abVCN
End Function

Public Function GetDataRunsString(ByVal mftRecNum As Long, attrType As Long, Optional ByVal drOffset As Integer = 0) As String
   Dim s As String
   Dim b As Byte, drSize As Byte
   Dim i As Long, j As Long, ret As Long, iOffset As Long
   Dim abRecord() As Byte
   
   ret = ReadMFTRecord(mftRecNum, abRecord)
   If ret <> 0 Then
      MsgBox ApiError(ret), vbCritical, "Error reading #" & mftRecNum & " MFT record"
      Exit Function
   End If
   ret = FixUpMFTRecord(abRecord)
   If ret <> 0 Then
      MsgBox ApiError(ret), vbCritical, "Error reading #" & mftRecNum & " MFT record"
      Exit Function
   End If
   iOffset = FindAttribute(attrType, abRecord)
   If iOffset <= 0 Then Exit Function
   If drOffset = 0 Then
      CopyMemory drOffset, abRecord(iOffset + 32), 2
   End If
   iOffset = iOffset + drOffset
   Do
     b = abRecord(iOffset + i)
     If b = 0 Then Exit Do
     drSize = b And &HF
     drOffset = (b And &HF0) / 16
     s = s & PadLeft(Hex(b), 2)
     For j = 1 To drSize
         b = abRecord(iOffset + i + j)
         s = s & PadLeft(Hex(b), 2)
     Next j
     For j = drSize + 1 To drSize + drOffset
         b = abRecord(iOffset + i + j)
         s = s & PadLeft(Hex(b), 2)
     Next j
     i = i + drSize + drOffset + 1
     If i > 255 Then Exit Do
   Loop
   If Right(s, 2) = "00" Then s = Left(s, Len(s) - 2)
   GetDataRunsString = s
End Function

Private Function ExtractDataRuns(ByRef adr As ATTR_DATA_RUNS, ByRef abData() As Byte) As Long
   If adr.realDataSize < 1 Then Exit Function
   ReDim abData(adr.allocatedDataSize - 1)
   Dim i As Long, ret As Long, bpc As Long, lastByte As Long, n As Long
   bpc = m_Info.BytesPerCluster
   Dim abTemp() As Byte
   For i = 0 To adr.drCount - 1
       With adr.dataRuns(i)
          If .nClusters > 0 Then
             If Not .bSparse Then
                If .LCN + .nClusters < m_Info.TotalClusters Then
                   n = .nClusters * bpc
                   ReDim abTemp(n - 1)
                   ret = DirectRead(.LCN * m_Info.SectorsPerCluster, 0, n, abTemp, False)
                   If ret = 0 Then
                      CopyMemory abData(lastByte), abTemp(0), n
                   Else
                      MsgBox ApiError(ret), vbCritical, "ExtractDataRuns Error"
                   End If
                   lastByte = lastByte + .nClusters * bpc
                End If
             End If
          End If
      End With
   Next i
   ReDim Preserve abData(adr.realDataSize - 1)
   Erase abTemp
End Function

Private Function MarkClusters(ByVal recNum As Long, adr As ATTR_DATA_RUNS) As Long
   Dim i As Long
   Dim cb As CLUSTERS_BLOCK
   If adr.drCount <= 0 Then Exit Function
   If clusterBlocksCount + adr.drCount > UBound(aClusters) Then ReDim Preserve aClusters(clusterBlocksCount + cbChunkSize - 1)
   
   cb.attrType = adr.attrType
   cb.mftRecordNum = recNum
   For i = 0 To adr.drCount - 1
       If adr.dataRuns(i).nClusters > 0 Then
          If Not adr.dataRuns(i).bSparse Then
             With adr.dataRuns(i)
                cb.startLCN = .LCN
                cb.endLCN = .LCN + .nClusters - 1
             End With
             aClusters(clusterBlocksCount) = cb
             clusterBlocksCount = clusterBlocksCount + 1
          End If
       End If
   Next i
End Function


'=====Get Full qualified path for MFT Record=====
'return path in format FullPath[(orphan,deleted)]:$AttribyteType
Public Function GetFullPath(ByVal recNum As Long) As String
   Dim s As String
   Dim bOrphan As Boolean
   Dim rec As MFT_RECORD_INFO, recParent As MFT_RECORD_INFO
   Dim numMain As Long, numParent As Long
   If aRecords(recNum).Type = rtEmpty Then
      s = "(Empty record)"
      GetFullPath = s
      Exit Function
   End If
   numMain = recNum
   Do While aRecords(numMain).mainRef.dwNumLow > 0
      numMain = aRecords(numMain).mainRef.dwNumLow
   Loop
   rec = aRecords(numMain)
   s = rec.FileName
   If rec.parentRef.dwNumLow > 5 Then
      If (aRecords(recNum).Type And rtOrphan) Then bOrphan = True
      numParent = rec.parentRef.dwNumLow
      Do While numParent > 5
         If bOrphan Then Exit Do
         If (aRecords(numParent).Type And rtOrphan) Then bOrphan = True
         s = aRecords(numParent).FileName & "\" & s
         numParent = aRecords(numParent).parentRef.dwNumLow
      Loop
   End If
   If bOrphan Then
      s = s & " (orphan)"
   Else
      If s = "." Then s = "$Root"
      If s = "" Then
         s = "==???==(MFT record #" & rec.nodeRef.dwNumLow
         If rec.mainRef.dwNumLow > 0 Then
            s = s & " - extension of record #" & rec.mainRef.dwNumLow
         End If
         s = s & ". Parent record #" & rec.parentRef.dwNumLow
         s = s & ")"
      End If
      s = DriveLetter & ":\" & s
      If (rec.Flags And 1) = 0 Then s = s & " (deleted)"
   End If
   GetFullPath = s
End Function

'=====Get Full qualified path for MFT Record=====
'Uses direct Drive access instead of aRecords() array
'return path in format FullPath[(orphan,deleted)]:$AttribyteType
Public Function GetFullPathDirect(ByVal mftNum As Long, ByRef sPath As String) As Long
    Dim abRecord() As Byte
    Dim ret As Long
    ret = ReadMFTRecord(mftNum, abRecord)
    If ret <> 0 Then Exit Function
    ret = FixUpMFTRecord(abRecord)
    If ret <> 0 Then Exit Function
    Dim offset As Long
    offset = FindAttribute(&H30, abRecord)
    If offset < 0 Then Exit Function
    Dim attrResident As ATTR_RESIDENT
    Dim attrFileName As ATTR_FILENAME
    CopyMemory attrResident, abRecord(offset + 16), Len(attrResident)
    CopyMemory attrFileName, abRecord(offset + attrResident.wAttrOffset), attrResident.dwLength
    If sPath = "" Then
       sPath = Left(StrConv(attrFileName.wFilename, vbFromUnicode), attrFileName.chFileNameLength)
    Else
       sPath = Left(StrConv(attrFileName.wFilename, vbFromUnicode), attrFileName.chFileNameLength) & "\" & sPath
    End If
    If attrFileName.dwMftParentDir.dwNumLow > 5 Then
       Call GetFullPathDirect(attrFileName.dwMftParentDir.dwNumLow, sPath)
    End If
End Function

'=====Search specific attribute for MFT record=====
'recNum - MFT record number
'searchAttribute - attribute tyep to search
'returns offset to attribute
Public Function FindRecordAttribute(ByVal recNum As Long, ByVal searchAttribute As Long) As Long
   Dim abRecord() As Byte
   Dim ret As Long
   ret = ReadMFTRecord(recNum, abRecord)
   If ret <> 0 Then
      FindRecordAttribute = -1
      MsgBox ApiError(ret), vbCritical, "Read mft record #" & recNum & "error."
      Exit Function
   End If
   FindRecordAttribute = FindAttribute(searchAttribute, abRecord)
End Function

Private Function FindAttribute(ByVal searchAttribute As Long, ByRef abRecord() As Byte, Optional recordOffset As Long = 0) As Long
    Dim offset As Integer
    Dim attrType As Long, dwSize As Long
    FindAttribute = -1
    If abRecord(recordOffset) <> &H46 Then Exit Function
    CopyMemory offset, abRecord(recordOffset + 20), 2
    CopyMemory attrType, abRecord(recordOffset + offset), 4
    If (attrType < 0) Or (attrType > searchAttribute) Then Exit Function
    Do While attrType < searchAttribute
       If attrType < 0 Then Exit Do
       CopyMemory dwSize, abRecord(recordOffset + offset + 4), 4
       offset = offset + dwSize
       CopyMemory attrType, abRecord(recordOffset + offset), 4
    Loop
    If (attrType < 0) Or (attrType > searchAttribute) Then Exit Function
    CopyMemory dwSize, abRecord(recordOffset + offset + 4), 4
    CopyMemory attrType, abRecord(recordOffset + offset + dwSize), 4
    If attrType = searchAttribute Then
       offset = offset + dwSize
    End If
    FindAttribute = offset
End Function

'=====Check $VOLUME:$BITMAP bit array=====
Public Function IsClusterBusy(ByVal LCN As Long) As Boolean
   Dim bClusterMap As Byte
   Dim iStart As Long, iRest As Long
   iStart = Int(LCN / 8)
   bClusterMap = aBitmap(iStart)
   If bClusterMap = 0 Then
      IsClusterBusy = False
   Else
      iRest = LCN - iStart * 8
      IsClusterBusy = IsBitSet(bClusterMap, iRest)
   End If
End Function

'=====Check $MFT:$BITMAP bit array=====
Public Function IsRecordBusy(ByVal recNum As Long) As Boolean
   Dim bRecordMap As Byte
   Dim iStart As Long, iRest As Long
   iStart = Int(recNum / 8)
   If iStart > UBound(abMFTBitmap) Then iStart = UBound(abMFTBitmap)
   bRecordMap = abMFTBitmap(iStart)
   If bRecordMap = 0 Then
      IsRecordBusy = False
   Else
      iRest = recNum - iStart * 8
      IsRecordBusy = IsBitSet(bRecordMap, iRest)
   End If
End Function

Private Function InitVolumeBitMap() As Long
   Dim abTemp() As Byte
   Dim q As Long
   Dim ret As Long, written As Long
   Dim startLCN As LARGE_INTEGER
   Dim BitmapSize As LARGE_INTEGER
   
   q = 1024# * 1024 * 64
   ReDim abTemp(q - 1)
   ret = DeviceIoControl(hDevice, FSCTL_GET_VOLUME_BITMAP, startLCN, Len(startLCN), abTemp(0), q, written, 0)
   If ret = 0 Then
      InitVolumeBitMap = ret
      Exit Function
   End If

   CopyMemory startLCN, abTemp(0), Len(startLCN)
   CopyMemory BitmapSize, abTemp(Len(startLCN)), Len(BitmapSize)
   If BitmapSize.LowPart > 0 Then
      If BitmapSize.LowPart Mod 8 Then 'convert bits to bytes
         q = BitmapSize.LowPart \ 8 - 1
      Else
         q = BitmapSize.LowPart \ 8
      End If
      ReDim aBitmap(q)
      CopyMemory aBitmap(0), abTemp(Len(startLCN) + Len(BitmapSize)), q
   End If
   Erase abTemp
End Function

Private Sub Class_Initialize()
    hDevice = INVALID_HANDLE_VALUE
    Set m_Info = New cVolumeInfo
    If memChunkSize = 0 Then
       memChunkSize = default_memChunkSize
       recChunkSize = default_recChunkSize
       cbChunkSize = default_cbChunkSize
    End If
    bInit = False
End Sub

Private Sub Class_Terminate()
    If hDevice <> INVALID_HANDLE_VALUE Then
       CloseHandle hDevice
       hDevice = INVALID_HANDLE_VALUE
    End If
'On Error Resume Next
    Erase aRecords
    Erase aClusters
    Erase aBitmap
    Erase abMFTBitmap
    Erase aDeleted
    Set m_Info = Nothing
End Sub

'*****User's interface staff*****
'=====Draw region of MFT records as colored blocks=====
'pb - picturebox for drawing
'blockSize - each block height and width
'rowBlocks - #blocks per 1 row
'startRecord, lastRecord - record's region to darw
Public Sub DrawMFTMap(pb As PictureBox, _
                      ByVal blockSize As Long, ByVal rowBlocks As Long, _
                      ByVal startRecord As Long, ByVal lastRecord As Long)
   If bScanning Then Exit Sub
   If Not bInit Then Exit Sub
   Dim record As MFT_RECORD_INFO
   Dim i As Long, j As Long, col As Long, row As Long, num As Long, clr As Long
   Dim recCount As Long
   recCount = m_Info.MFTRecordsCount
   pb.Cls
   For i = startRecord To lastRecord
      If i > recCount - 1 Then Exit For
      record = aRecords(i)
      num = record.nodeRef.dwNumLow
      clr = vbBlue
      j = i - startRecord
      row = Int(j / rowBlocks)
      col = j - row * rowBlocks
      If record.Type = rtEmpty Then
         clr = vbWhite
      Else
         If (record.Type And rtInvalid) Then
            clr = vbMagenta
         Else
            If record.mainRef.dwNumLow > 0 Then
               clr = vbCyan
            Else
               If record.Type And rtOrphan Then
                  clr = colOck
               Else
                  If (record.Flags And &H1) = 0 Then
                     If (record.Flags And &H2) = 0 Then clr = colL3Green Else clr = colLGreen
                  Else
                     If (record.Flags And &H2) = 0 Then clr = colL2Blue Else clr = colBlue
                  End If
               End If
            End If
         End If
      End If
      pb.Line (col * blockSize, row * blockSize)-(col * blockSize + blockSize - 2, row * blockSize + blockSize - 2), clr, BF
   Next i
End Sub

'=====Draw region of clusters as colored blocks=====
'pb - picturebox for drawing
'blockSize - each block height and width
'rowBlocks - #blocks per 1 row
'startCluster, lastCluster - cluster's region to darw
Public Sub DrawClustersMap(pb As PictureBox, _
                           ByVal blockSize As Long, ByVal rowBlocks As Long, _
                           ByVal startCluster As Long, ByVal endCluster As Long)
   If bScanning Then Exit Sub
   If Not bInit Then Exit Sub
   Dim i As Long, clusterNum As Long, pos As Long
   Dim col As Long, row As Long, num As Long, clr As Long
   Dim bFound As Boolean
   Dim cb As CLUSTERS_BLOCK, cbTemp As CLUSTERS_BLOCK
   pb.Cls
'sanitize
   If startCluster < 0 Then startCluster = 0
   If endCluster > m_Info.TotalClusters - 1 Then endCluster = m_Info.TotalClusters
'search each cluster within CLUSTERS_BLOCKS array
   For clusterNum = startCluster To endCluster
       num = clusterNum - startCluster
       row = Int(num / rowBlocks)
       col = num - row * rowBlocks
       pos = 0: bFound = False
       Call SearchClusterBlocks(pos, bFound, clusterNum, 0, clusterBlocksCount - 1)
       If Not bFound Then 'no one CLUSTERS_BLOCKS starts from clusterNum
          pos = pos - 1
          If pos < 0 Then pos = 0
          cb = aClusters(pos)
          If clusterNum >= cb.startLCN Then 'check previous record
             If clusterNum <= cb.endLCN Then
                bFound = True
             End If
          End If
          If Not bFound Then 'check cross-links
             For i = 0 To cb.crossLinksCount - 1
                 cbTemp = aClusters(cb.crossLinkIndex(i))
                 If clusterNum >= cbTemp.startLCN Then
                    If clusterNum <= cbTemp.endLCN Then
                       bFound = True
                       cb = cbTemp
                       Exit For
                    End If
                 End If
             Next i
          End If
       Else
          cb = aClusters(pos) 'found CLUSTERS_BLOCK starts from clusterNum
       End If
       If Not bFound Then
          If (clusterNum >= m_Info.MFTStartLcn) And (clusterNum <= m_Info.MFTZoneEnd) Then
              If clusterNum > m_Info.MFTZoneStart Then
                 clr = colL3Green 'unused MFT area
              Else
                 num = ((clusterNum - m_Info.MFTStartLcn) * m_Info.BytesPerCluster) \ m_Info.BytesPerMFTRecord
                 If IsRecordBusy(num) Then
                    clr = colDGreen 'occupied MFT record
                  Else
                    clr = colLGreen 'Free MFT Record
                  End If
              End If
          ElseIf (clusterNum >= m_Info.MFT2StartLcn) And (clusterNum <= m_Info.MFT2StartLcn + 4 * (4 * m_Info.BytesPerMFTRecord) \ m_Info.BytesPerCluster) Then
               clr = colDGreen 'occupied MFT Mirror record
          Else
             If IsClusterBusy(clusterNum) Then
                clr = vbCyan 'Not mapped by MFT (not found when search), but marked as occupied by $MFT:$BITMAP
             Else
                clr = vbWhite 'Free
             End If
          End If
       Else
          If cb.crossLinksCount > 0 Then
             clr = vbMagenta ' cross-linked
          Else
             clr = vbBlue 'normal, occupied by 1 record
          End If
       End If
       pb.Line (col * blockSize, row * blockSize)-(col * blockSize + blockSize - 2, row * blockSize + blockSize - 2), clr, BF
   Next clusterNum
End Sub

'=====Get string information for MFT record====
'mftNum - MFT record number
Public Function GetMFTRecordInfoShort(ByVal mftNum As Long) As String
   If bScanning Then Exit Function
   If mftNum > m_Info.MFTRecordsCount - 1 Then Exit Function
   Dim s As String, sPath As String
   Dim numMain As Long
   
   s = "Record #" & mftNum
   sPath = aRecords(mftNum).FileName
   If aRecords(mftNum).Type = rtEmpty Then
      If sPath = "" Then sPath = "(Empty record)"
      s = s & vbCrLf & "FileName: " & sPath
      GetMFTRecordInfoShort = s
      Exit Function
   End If
      
   If aRecords(mftNum).mainRef.dwNumLow > 0 Then
      numMain = aRecords(mftNum).mainRef.dwNumLow
      s = s & " (extended record)." & vbCrLf & "Main record: #" & numMain
      'numMain = mftNum
      'Do While aRecords(numMain).mainRef.dwNumLow > 0
      '   s = s & "#" & numMain & ","
      '   numMain = aRecords(numMain).mainRef.dwNumLow
      'Loop
      If Right(s, 1) = "," Then s = Left(s, Len(s) - 1)
      s = s & vbCrLf & "Main record file path: " & GetFullPath(numMain)
      GetMFTRecordInfoShort = s
      Exit Function
   End If
   
   s = s & vbCrLf & "File path: " & GetFullPath(mftNum)
   s = s & vbCrLf & "Record marked as " & IIf(IsRecordBusy(mftNum), "Occupied", "Free")
   
   GetMFTRecordInfoShort = s
End Function

'=====Get string information for MFT record====
'clusterNum - cluster number
Public Function GetClusterInfoShort(ByVal clusterNum As Long) As String
   If bScanning Then Exit Function
   If clusterNum > m_Info.TotalClusters - 1 Then Exit Function
   Dim s As String, sPath As String
   Dim i As Long, num As Long, pos As Long
   Dim cb As CLUSTERS_BLOCK, cbTemp As CLUSTERS_BLOCK
   Dim bFound As Boolean
       
   Call SearchClusterBlocks(pos, bFound, clusterNum, 0, clusterBlocksCount - 1)
   If Not bFound Then
      pos = pos - 1
      If pos < 0 Then pos = 0
      cb = aClusters(pos)
      If clusterNum >= cb.startLCN Then
         If clusterNum <= cb.endLCN Then bFound = True
      End If
      If Not bFound Then
         For i = 0 To cb.crossLinksCount - 1
             cbTemp = aClusters(cb.crossLinkIndex(i))
             If clusterNum >= cbTemp.startLCN Then
                If clusterNum <= cbTemp.endLCN Then
                   bFound = True
                   cb = cbTemp
                   Exit For
                End If
             End If
         Next i
      End If
   Else
      cb = aClusters(pos)
   End If
  
   s = "Cluster #" & clusterNum
   If Not bFound Then
      If (clusterNum >= m_Info.MFTStartLcn) And (clusterNum <= m_Info.MFTZoneEnd) Then
         If clusterNum > m_Info.MFTZoneStart Then
            s = s & " - Reserved MFT Area"
         Else
            num = ((clusterNum - m_Info.MFTStartLcn) * m_Info.BytesPerCluster) \ m_Info.BytesPerMFTRecord
            s = s & " - MFT record(s) #" & num & " - #" & num + m_Info.BytesPerCluster \ m_Info.BytesPerMFTRecord
            If IsRecordBusy(num) Then
               s = s & " (occupied)"
            Else
               s = s & " (free)"
            End If
         End If
      ElseIf (clusterNum >= m_Info.MFT2StartLcn) And (clusterNum <= m_Info.MFT2StartLcn + (4 * m_Info.BytesPerMFTRecord) \ m_Info.BytesPerCluster) Then
         s = s & " - MFT Mirror area"
      Else
         If IsClusterBusy(clusterNum) Then
            s = s & " - Not MFT mapped (system?)"
         Else
            s = s & " - Free cluster"
         End If
      End If
   Else
      If cb.crossLinksCount > 0 Then
          s = s & " - Cross linked"
      End If
      sPath = GetFullPath(cb.mftRecordNum)
      s = s & vbCrLf & sPath & ":$" & TranslateAttrType(cb.attrType)
      For i = 0 To cb.crossLinksCount - 1
          num = aClusters(cb.crossLinkIndex(i)).mftRecordNum
          sPath = GetFullPath(num)
          s = s & vbCrLf & sPath & ":$" & TranslateAttrType(aClusters(cb.crossLinkIndex(i)).attrType)
      Next i
   End If
   GetClusterInfoShort = s
End Function

'=====Fill TreeView with root records (first 27 MFT records)=====
Public Sub AddRootNodes(ByVal tv As Object, Optional ByVal recFilter As RECORDS_TREE_FILTER = rtfUndefined)
   RaiseEvent StatusChanged("Populating record's tree...")
   Dim nd As Object
   tv.Nodes.Clear
   If recFilter = rtfUndefined Then recFilter = m_Filter
   If recFilter <> rtfDeleted Then
      Set nd = tv.Nodes.Add(, , "Orphans", " Orphans", "folderClosedHidden")
      nd.SelectedImage = nd.Image
      nd.ExpandedImage = "folderOpenHidden"
      nd.Tag = "Orphans"
      EnumOrphans tv
   End If

   If recFilter > rtfAllRecords Then
      If recFilter <> rtfOrphans Then
         Set nd = tv.Nodes.Add(, , "rec_" & 5, " $Root (" & DriveLetter & ":\)", "drive", "drive")
         nd.Tag = 5
         tv.Nodes.Add(nd, tvwChild, , "Dummy_node").Tag = "Dummy_Node"
      End If
      RaiseEvent StatusChanged("Done")
      Exit Sub
   End If
   
   Dim i As Long, num As Long
   Dim s As String
   For i = 0 To 27
       With aRecords(i)
          num = .nodeRef.dwNumLow
          If (num < 27) And (.parentRef.dwNumLow = 5) Then
             s = .FileName
             If Len(s) > 0 Then
                If s = "." Then s = " $Root (" & DriveLetter & ":)"
                Set nd = tv.Nodes.Add(, , "rec_" & num, s)
                nd.Tag = num
                If .Flags And 2 Then
                   If .FileName = "." Then
                      nd.Image = "drive"
                      nd.ExpandedImage = "drive"
                   Else
                      nd.Image = "folderClosedHidden"
                      nd.ExpandedImage = "folderOpenHidden"
                   End If
                   nd.SelectedImage = nd.Image
                   tv.Nodes.Add(nd, tvwChild, , "Dummy_node").Tag = "Dummy_Node"
                Else
                   If .Flags And 1 Then
                      nd.Image = "fileHidden"
                   Else
                      nd.Image = "delete"
                   End If
                   nd.SelectedImage = nd.Image
                   nd.ExpandedImage = nd.Image
                End If
             End If
          End If
       End With
   Next i
   RaiseEvent StatusChanged("Done")
End Sub

'=====Fill TreeView node with child records =====
Public Sub EnumNodesUnder(ByVal tv As Object, ByVal ndParent As Object, Optional ByVal recFilter As RECORDS_TREE_FILTER = rtfUndefined)
   If recFilter = rtfUndefined Then recFilter = m_Filter
   If (recFilter And rtfDeleted) Then
      EnumDeletedCildren tv, ndParent
      Exit Sub
   End If
   Dim i As Long, num As Long, numParent As Long
   Dim nd As Object
   Dim s As String
   numParent = ndParent.Tag
   For i = 0 To UBound(aRecords)
       With aRecords(i)
           If .Type And rtOrphan Then GoTo SkipRecord
           If .mainRef.dwNumLow > 0 Then GoTo SkipRecord
           If .parentRef.dwNumLow = numParent Then
              s = .FileName
              If s = "" Then s = "==???==_" & .Type
              num = .nodeRef.dwNumLow
              Set nd = Nothing
              On Error Resume Next
              Set nd = tv.Nodes.Add(ndParent, tvwChild, "rec_" & num, s)
              On Error GoTo 0
              If Not nd Is Nothing Then
                 nd.Tag = num
                 If .Flags And 2 Then
                    If .Flags And 1 Then
                       nd.Image = "folderClosed"
                       nd.ExpandedImage = "folderOpen"
                    Else
                       nd.Image = "folderDeleted"
                       nd.ExpandedImage = nd.Image
                    End If
                    nd.SelectedImage = nd.Image
                    tv.Nodes.Add(nd, tvwChild, , "Dummy_node").Tag = "Dummy_Node"
                    nd.Text = " " & nd.Text
                 Else
                    If .Flags And 1 Then
                       If num < 27 Then nd.Image = "fileHidden" Else nd.Image = "fileGeneric"
                    Else
                       nd.Image = "delete"
                    End If
                    nd.SelectedImage = nd.Image
                    nd.ExpandedImage = nd.Image
                 End If
              End If
           End If
SkipRecord:
       End With
   Next i
   ndParent.Sorted = True
End Sub

'=====Fill TreeView with ALL deleted records =====
Public Sub EnumAllDeleted(tv As Object, rootNode As Object)
   Dim i As Long, j As Long, num As Long, numParent As Long
   Dim bFound As Boolean
   Dim nd As Object, ndParent As Object
   Dim s As String
   Dim rec As MFT_RECORD_INFO
   Dim dicParents As New Dictionary
   Dim colTemp As Collection
   dicParents.Add 5, "rec_5"
   For i = 27 To UBound(aRecords)
       With aRecords(i)
           If .mainRef.dwNumLow > 0 Then GoTo SkipRecord
           If .Type And rtOrphan Then GoTo SkipRecord
           If (.Flags And 1) = 0 Then 'deleted
              numParent = .parentRef.dwNumLow
              Set colTemp = New Collection
              Do While numParent >= 5
                 If dicParents.Exists(numParent) Then Exit Do
                 colTemp.Add numParent
                 numParent = aRecords(numParent).parentRef.dwNumLow
              Loop
              Set ndParent = tv.Nodes("rec_" & numParent)
              If colTemp.Count > 0 Then
                 For j = colTemp.Count To 1 Step -1
                     num = colTemp(j)
                     If aRecords(num).Flags And 2 Then
                         Set ndParent = tv.Nodes.Add(ndParent, tvwChild, "rec_" & num, aRecords(num).FileName)
                     If Err.Number Then
                        Debug.Print num, aRecords(num).FileName
                        Err.Clear
                        GoTo SkipRecord
                     End If
                        If aRecords(num).Flags And 1 Then 'In use
                           ndParent.Image = "folderClosed"
                           ndParent.ExpandedImage = "folderOpen"
                        Else 'Deleted
                           ndParent.Image = "folderDeleted"
                           ndParent.ExpandedImage = ndParent.Image
                        End If
                        ndParent.SelectedImage = ndParent.Image
                        dicParents.Add num, "rec_" & num
                    End If
                 Next j
              End If
'              DoEvents
              num = .nodeRef.dwNumLow
              If .Flags And 2 Then 'Folder
                 If Not dicParents.Exists(num) Then
                    Set nd = tv.Nodes.Add(ndParent, tvwChild, "rec_" & num, .FileName, "folderDeleted")
                    nd.SelectedImage = nd.Image
                    nd.Tag = num
                    dicParents.Add num, "rec_" & num
                 End If
              Else 'File
                 Set nd = tv.Nodes.Add(ndParent, tvwChild, "rec_" & num, .FileName, "delete")
                 nd.SelectedImage = nd.Image
                 nd.Tag = num
              End If
           End If
SkipRecord:
       End With
   Next i
   Set dicParents = Nothing
   Set colTemp = Nothing
'   ndParent.Sorted = True
End Sub

Public Sub EnumDeletedCildren1(tv As Object, ByVal ndParent As Object)
   Dim i As Long, num As Long, numParent As Long
   Dim bFound As Boolean
   Dim nd As Object
   Dim s As String
   numParent = ndParent.Tag
   For i = 0 To aDeleted(0).Children - 1
      Debug.Print aDeleted(0).Child(i)
      tv.Nodes.Add ndParent, tvwChild, , aRecords(aDeleted(0).Child(i)).FileName
   Next i
End Sub

'=====Fill TreeView node with deleted child records =====
Public Sub EnumDeletedCildren(tv As Object, ByVal ndParent As Object)
   Dim i As Long, num As Long, numParent As Long
   Dim bFound As Boolean
   Dim nd As Object
   Dim s As String
   numParent = ndParent.Tag
   For i = 27 To UBound(aRecords)
       With aRecords(i)
           If .mainRef.dwNumLow > 0 Then GoTo SkipRecord
           If .Type And rtOrphan Then GoTo SkipRecord
           If .parentRef.dwNumLow = numParent Then
              s = .FileName
              If s = "" Then s = "==???==_" & .Type
              num = .nodeRef.dwNumLow
              If .Flags And 2 Then 'Folder
                 bFound = False
                 Call FindDeletedChildren(num, bFound)
                 If .Flags And 1 Then 'InUse
                    If bFound Then
                       Set nd = tv.Nodes.Add(ndParent, tvwChild, "rec_" & num, s, "folderClosed")
                       nd.ExpandedImage = "folderOpen"
                       nd.SelectedImage = nd.Image
                       nd.Tag = num
                       tv.Nodes.Add(nd, tvwChild, , "Dummy_node").Tag = "Dummy_Node"
                       nd.Text = " " & nd.Text
                    End If
                 Else 'Deleted
                    Set nd = tv.Nodes.Add(ndParent, tvwChild, "rec_" & num, s, "folderDeleted")
                    nd.SelectedImage = nd.Image
                    nd.Tag = num
                    If bFound Then
                       tv.Nodes.Add(nd, tvwChild, , "Dummy_node").Tag = "Dummy_Node"
                       nd.Text = " " & nd.Text
                    End If
                 End If
              Else 'File
                 If (.Flags And 1) = 0 Then 'Deleted
                    Set nd = tv.Nodes.Add(ndParent, tvwChild, "rec_" & num, s, "delete")
                    nd.SelectedImage = nd.Image
                    nd.Tag = num
                 End If
              End If
           End If
SkipRecord:
       End With
       If (i Mod 1000) = 0 Then
          RaiseEvent Progress(i / UBound(aRecords) * 100)
       End If
   Next i
   ndParent.Sorted = True
End Sub

Private Sub FindDeletedChildren(ByVal numParent As Long, ByRef bFound As Boolean)
   Dim i As Long
   For i = 27 To UBound(aRecords)
       If i = numParent Then GoTo SkipRecord
       With aRecords(i)
          If .Type And rtOrphan Then GoTo SkipRecord
          If (.Flags And 1) = 0 Then 'deleted
             If IsChild(numParent, .nodeRef.dwNumLow) Then
                bFound = True
                Exit For
             End If
          End If
SkipRecord:
       End With
   Next i
End Sub

Private Function IsChild(ByVal numParent As Long, ByVal numChild As Long) As Boolean
   Dim i As Long, nextParent As Long
   nextParent = aRecords(numChild).parentRef.dwNumLow
   If nextParent = numParent Then
      IsChild = True
      Exit Function
   End If
   Do While nextParent > 5
      nextParent = aRecords(nextParent).parentRef.dwNumLow
      If nextParent = numParent Then
         IsChild = True
         Exit Do
      End If
   Loop
End Function

'=====Enumerate dleted files=====
Public Sub EnumDeletedParents()
    ReDim aDeleted(default_delParents - 1)
    Dim dictDeleted As New Dictionary
    Dim i As Long, j As Long, num As Long, numParent As Long, idx As Long, nParents As Long
    Dim colParents As Collection
    dictDeleted.Add 5, 0
    deletedParentsCount = 1
    For i = 0 To UBound(aRecords)
       If deletedParentsCount > UBound(aDeleted) - 4 Then
           ReDim Preserve aDeleted(UBound(aDeleted) + default_delParents)
       End If
       With aRecords(i)
           If (.Flags And 1) = 0 Then
              If .Type And rtOrphan Then GoTo SkipRecord
              If .mainRef.dwNumLow > 0 Then GoTo SkipRecord
              num = .nodeRef.dwNumLow
              If num = 0 Then GoTo SkipRecord
              numParent = .parentRef.dwNumLow
              If numParent = 0 Then numParent = 5
              Set colParents = New Collection
              Do While numParent > 5
                 colParents.Add numParent
                 numParent = aRecords(numParent).parentRef.dwNumLow
              Loop
              colParents.Add numParent
              For j = colParents.Count To 2 Step -1
                 If Not dictDeleted.Exists(colParents(j)) Then
                    dictDeleted.Add colParents(j), deletedParentsCount
                    deletedParentsCount = deletedParentsCount + 1
                 Else
                    If Not dictDeleted.Exists(colParents(j - 1)) Then
                       dictDeleted.Add colParents(j - 1), deletedParentsCount
                       deletedParentsCount = deletedParentsCount + 1
                       idx = dictDeleted(colParents(j))
                       With aDeleted(idx)
                           ReDim Preserve .Child(.Children)
                           .Child(.Children) = colParents(j - 1)
                           .Children = .Children + 1
                       End With
                    End If
                 End If
              Next j
              If aRecords(num).Flags And 2 Then
                 If Not dictDeleted.Exists(num) Then
                    dictDeleted.Add num, deletedParentsCount
                    deletedParentsCount = deletedParentsCount + 1
                 End If
              End If
              idx = dictDeleted(colParents(1))
              With aDeleted(idx)
                   ReDim Preserve .Child(.Children)
                   .Child(.Children) = num
                   .Children = .Children + 1
              End With
           End If
SkipRecord:
       End With
       If (i Mod 1000) = 0 Then
          RaiseEvent Progress(i / UBound(aRecords) * 100)
       End If
    Next i
    ReDim Preserve aDeleted(deletedParentsCount - 1)
    Set colParents = Nothing
    Set dictDeleted = Nothing
End Sub

'=====Enumerate orphan dleted files=====
Public Sub EnumOrphans(tv As Object, Optional ByVal strImageKey As String = "tableDeleted")
   Dim i As Long, num As Long
   Dim nd As Object, ndParent As Object
   Set ndParent = tv.Nodes("Orphans")
   If ndParent Is Nothing Then Exit Sub
   RaiseEvent StatusChanged("Enumerating orphans...")
   For i = 0 To UBound(aRecords)
       With aRecords(i)
           If .Type And rtOrphan Then
              num = .nodeRef.dwNumLow
              Set nd = tv.Nodes.Add(ndParent, tvwChild, "rec_" & num, .FileName, strImageKey)
              nd.SelectedImage = nd.Image
              nd.Tag = num
           End If
       End With
       If (i Mod 1000) = 0 Then
          RaiseEvent Progress(i / UBound(aRecords) * 100)
       End If
   Next i
   RaiseEvent Progress(100)
   RaiseEvent Progress(0)
   RaiseEvent StatusChanged("Sorting orphans...")
   ndParent.Sorted = True
   RaiseEvent StatusChanged("Done")
End Sub

'=====Show HexDump====
Public Function ShowDump(ByVal iStartSector As Long, ByVal nBytes As Long, ByRef grd As MSFlexGrid) As Long
   Dim abData() As Byte
   Dim ret As Long, offset As Long
   Dim strASCII As String
   Dim b As Byte
   
   ret = DirectRead(iStartSector, 0, nBytes, abData, False)
   If ret Then
      Erase abData
      ShowDump = ret
      Exit Function
   End If
   Dim row As Long, col As Long, i As Long
   For i = 0 To UBound(abData)
       row = Int(i / 16)
       If row > grd.Rows Then Exit For
       col = i - row * 16
       b = abData(i)
       If b > 31 Then strASCII = strASCII & Chr(b) Else strASCII = strASCII & Chr(&H95) '"."
       If col = 15 Then
          grd.TextMatrix(row + 1, 17) = strASCII
          strASCII = ""
       End If
       grd.TextMatrix(row + 1, col + 1) = Right("0" & Hex(b), 2)
   Next i
   Erase abData
End Function

Public Sub SearchClusterBlocks(ByRef pos As Long, ByRef bFound As Boolean, _
                                ByVal searchLCN As Long, ByVal startPos As Long, _
                                ByVal endPos As Long)
    If searchLCN = 0 Then
       pos = 0
       Exit Sub
    End If
    If bFound Then Exit Sub
    Dim midPoint As Long
    midPoint = (startPos + endPos) \ 2
    pos = endPos: bFound = False
    If searchLCN = aClusters(midPoint).startLCN Then
       pos = midPoint
       bFound = True
    ElseIf endPos > startPos Then
       If searchLCN < aClusters(midPoint).startLCN Then
          Call SearchClusterBlocks(pos, bFound, searchLCN, startPos, midPoint - 1)
       Else
          Call SearchClusterBlocks(pos, bFound, searchLCN, midPoint + 1, endPos)
       End If
    Else
       If aClusters(midPoint).startLCN < searchLCN Then pos = endPos + 1
    End If
End Sub

'=====QuickSort aClusters array=====
Private Sub SortClustersBlocks(Optional ByVal plngLeft As Long, Optional ByVal plngRight As Long)
    Dim lngFirst As Long, lngLast As Long
    Dim midLCN As Long
    Static n As Long
    Dim cbSwap As CLUSTERS_BLOCK
    
    If plngRight = 0 Then
        plngLeft = 0
        plngRight = clusterBlocksCount - 1
        midLCN = aClusters((plngLeft + plngRight) \ 2).startLCN
        Do While midLCN = 0
           midLCN = aClusters(GetRandom(0, plngRight)).startLCN
        Loop
    End If
    lngFirst = plngLeft
    lngLast = plngRight
    If midLCN = 0 Then midLCN = aClusters((plngLeft + plngRight) \ 2).startLCN
    Do
        Do While aClusters(lngFirst).startLCN < midLCN And lngFirst < plngRight
            lngFirst = lngFirst + 1
        Loop
        Do While midLCN < aClusters(lngLast).startLCN And lngLast > plngLeft
            lngLast = lngLast - 1
        Loop
        If lngFirst <= lngLast Then
            cbSwap = aClusters(lngFirst)
            aClusters(lngFirst) = aClusters(lngLast)
            aClusters(lngLast) = cbSwap
            lngFirst = lngFirst + 1
            lngLast = lngLast - 1
        End If
    Loop Until lngFirst > lngLast
    n = n + 1
    If (n Mod 1000) = 0 Then
        RaiseEvent Progress(lngLast / clusterBlocksCount * 100)
    End If
    If plngLeft < lngLast Then SortClustersBlocks plngLeft, lngLast
    If lngFirst < plngRight Then SortClustersBlocks lngFirst, plngRight
End Sub

Private Sub CheckCrossLinks()
   Dim i As Long, j As Long
   For i = 0 To clusterBlocksCount - 2
       aClusters(i).blockIndex = i
       aClusters(i + 1).blockIndex = i + 1
       If aClusters(i + 1).startLCN <= aClusters(i).endLCN Then
          With aClusters(i)
             ReDim Preserve .crossLinkIndex(.crossLinksCount)
             .crossLinkIndex(.crossLinksCount) = i + 1
             .crossLinksCount = .crossLinksCount + 1
          End With
          With aClusters(i + 1)
             ReDim Preserve .crossLinkIndex(.crossLinksCount)
             .crossLinkIndex(.crossLinksCount) = i
             .crossLinksCount = .crossLinksCount + 1
          End With
       End If
       For j = 0 To aClusters(i).crossLinksCount - 1
           If aClusters(i + 1).blockIndex <> aClusters(i).crossLinkIndex(j) Then
              If aClusters(i + 1).startLCN >= aClusters(aClusters(i).crossLinkIndex(j)).startLCN Then
                 If aClusters(i + 1).startLCN <= aClusters(aClusters(i).crossLinkIndex(j)).endLCN Then
                    With aClusters(i + 1)
                       ReDim Preserve .crossLinkIndex(.crossLinksCount)
                       .crossLinkIndex(.crossLinksCount) = aClusters(i).crossLinkIndex(j)
                       .crossLinksCount = .crossLinksCount + 1
                    End With
              
                    With aClusters(aClusters(i).crossLinkIndex(j))
                       ReDim Preserve .crossLinkIndex(.crossLinksCount)
                       .crossLinkIndex(.crossLinksCount) = aClusters(i + 1).blockIndex
                       .crossLinksCount = .crossLinksCount + 1
                    End With
                 End If
              End If
           End If
       Next j
       RaiseEvent Progress(i / clusterBlocksCount * 100)
   Next i
   RaiseEvent Progress(100)
End Sub
