Attribute VB_Name = "mMFT"
Option Explicit

Public Type NTFS_RECORD_HEADER
  dwType As Long                    ' File type, for example 'FILE'
  wUsaOffset As Integer             ' Offset to the Update Sequence Array
  wUsaCount As Integer              ' Size in words of Update Sequence Array
  Lsn As LARGE_INTEGER              ' $LogFile Sequence Number (LSN)
End Type

Public Type FILE_RECORD_HEADER
  RecHdr As NTFS_RECORD_HEADER
  wSequenceNumber As Integer        ' Sequence number
  wLinkCount As Integer             ' Hard link count
  wAttributeOffset As Integer       ' Offset to the first Attribute
  wFlags As Integer                 ' Flags. bit 1 = in use, bit 2 = directory, bit 4 & 8 = unknown.
  dwBytesInUse As Long              ' Real size of the FILE record
  dwBytesAllocated As Long          ' Allocated size of the FILE record
  BaseFileRecord As INODE_REFERENCE ' File reference to the base FILE record
  wNextAttributeNumber As Integer   ' Next Attribute Id
  wPadding As Integer               ' Align to 4 UCHAR boundary (XP)
  dwMFTRecordNumber As Long         ' Number of this MFT Record (XP)
  wUpdateSeqNum As Integer          '
End Type

Public Type ATTR_HEADER
   dwType As Long
   dwFullLength As Long
   uchNonResFlag As Byte
   uchNameLength As Byte
   wNameOffset As Integer
   wFlags As Integer
   wID As Integer
End Type
  
Public Type ATTR_RESIDENT
   dwLength As Long
   wAttrOffset As Integer
   uchIndexedTag As Byte
   uchPadding As Byte
End Type

Public Type ATTR_NONRESIDENT
   n64StartVCN As LARGE_INTEGER
   n64EndVCN As LARGE_INTEGER
   wDatarunOffset As Integer
   wCompressionSize As Integer
   uchPadding(3) As Byte
   n64AllocSize As LARGE_INTEGER
   n64RealSize As LARGE_INTEGER
   n64StreamSize As LARGE_INTEGER
End Type

Public Type ATTR_STANDARD
   n64Create As LARGE_INTEGER
   n64Modify As LARGE_INTEGER
   n64Modfil As LARGE_INTEGER
   n64Access As LARGE_INTEGER
   dwFATAttributes As Long
   dwMaxVersions As Long
   dwVersionNum As Long
   dwClassID As Long
   dwOwnerID As Long
   dwSecurityID As Long
   n64QuotaCharged As LARGE_INTEGER
   n64FixupSeqNum As LARGE_INTEGER
End Type

Public Type ATTR_FILENAME
   dwMftParentDir As INODE_REFERENCE
   n64Create      As LARGE_INTEGER
   n64Modify      As LARGE_INTEGER
   n64Modfil      As LARGE_INTEGER
   n64Access      As LARGE_INTEGER
   n64Allocated   As LARGE_INTEGER
   n64RealSize    As LARGE_INTEGER
   dwFlags        As Long
   dwSecurityID   As Long
   chFileNameLength As Byte
   chFileNameType   As Byte
   'wFilename(511) As Byte
   wFilename As String * 512
End Type

Public Type MFT_RECORD_INFO
   nodeRef As INODE_REFERENCE
   mainRef As INODE_REFERENCE
   parentRef As INODE_REFERENCE
   Flags As Byte
   Type As Byte
   FileName As String
End Type

Public Type DATA_RUN
   vcn As Variant
   LCN As Variant
   nClusters As Variant
   bSparse As Boolean
End Type

Public Type ATTR_DATA_RUNS
   attrType As Long
   drCount As Long
   StartVCN As Variant
   EndVCN As Variant
   realDataSize As Variant
   allocatedDataSize As Variant
   streamSize As Variant
   dataRuns() As DATA_RUN
End Type

Public Type MFT_VALUE_INFO
   mviCaption As String
   mviDescription As String
   mviValue As String
   mviOffset As Long
   mviSize As Long
End Type

Public Type MFT_ATTR_INFO
   attrName As MFT_VALUE_INFO
   attrValues() As MFT_VALUE_INFO
   attrDataRuns As ATTR_DATA_RUNS
   attrResidentData() As Byte
End Type

Public Type MFT_RECORD_INFO_FULL
   recordHeader As MFT_ATTR_INFO
   recordAttributes() As MFT_ATTR_INFO
End Type

Public Type DATA_RUN_INFO
   LCN As Long
   nClusters As Long
   mftRecordNum As Long
   attrType As Long
End Type

Public currentVolume As cVolume
Public searchData As SEARCH_DATA
Public owData As OVERWRITE_DATA
   
   
Private abRecord() As Byte

Public Function GetMFTRecordInfoFull(ByVal mftNum As Long, ByRef info As MFT_RECORD_INFO_FULL) As Long
   If currentVolume Is Nothing Then
      GetMFTRecordInfoFull = 1411
      Erase abRecord
      Exit Function
   End If
   Dim offset As Long, size As Long, ret As Long
   ret = currentVolume.ReadMFTRecord(mftNum, abRecord)
   If ret Then
      GetMFTRecordInfoFull = ret
      Erase abRecord
      Exit Function
   End If
'******** fixup*******************
   ret = currentVolume.FixUpMFTRecord(abRecord)
   If ret Then
      GetMFTRecordInfoFull = ret
      Erase abRecord
      Exit Function
   End If
   CopyMemory size, abRecord(6), 2
'********Record header**************
   Dim abTemp() As Byte
   Dim s As String, s1 As String
   Dim attrOffset As Long, attrSize, attrDataSize As Long, i As Long
   Dim nodeRef As INODE_REFERENCE
   With info.recordHeader
       ReDim .attrValues(15)
       CopyMemory attrOffset, abRecord(20), 2
       SetMFTValueDirect .attrName, "MFT Record Header", "general info", "", 0, 48 + size * 2
       ReDim abTemp(3)
       CopyMemory abTemp(0), abRecord(0), 4
       s = StrConv(abTemp, vbUnicode)
       SetMFTValueDirect .attrValues(0), "Magic marker", "&H46494C45 (word ""FILE"" ASCII)", s, 0, 4
       SetMFTValueInfo .attrValues(1), "Fixup offset", "offset to update sequence", 4, 2
       SetMFTValueInfo .attrValues(2), "Fixup size", "Size in words of Update Sequence Number & Array (S)", 6, 2
       SetMFTValueInfo .attrValues(3), "$LogFile Sequence Number", "LSN", 8, 8
       SetMFTValueInfo .attrValues(4), "Sequence number", "Is increased by one each time the MFT record is reused. Used as a consistency check in the index buffers.", 16, 2
       SetMFTValueInfo .attrValues(5), "Hard links count", "", 18, 2
       SetMFTValueInfo .attrValues(6), "Offset to attributes bytes", "", 20, 2
       SetMFTValueInfo .attrValues(7), "Flags", "0-not in use record (deleted), 1-in use, 2-record is directory", 22, 2
       SetMFTValueInfo .attrValues(8), "Length of in-use part of the record", "Header + attributes. Rounded up to 8 bytes", 24, 4
       SetMFTValueInfo .attrValues(9), "Allocated length", "Total length of the record", 28, 4
       SetMFTValueInfo .attrValues(10), "File reference to main MFT record", "0 for base MFT file record", 32, 6
       SetMFTValueInfo .attrValues(11), "Next available attribute ID", "", 40, 2
       SetMFTValueInfo .attrValues(12), "Align to 4 byte boundary", "XP OS", 42, 2
       SetMFTValueInfo .attrValues(13), "Number of this MFT Record", "XP OS", 44, 4
       If size > 0 Then
          SetMFTValueInfo .attrValues(14), "Update Sequence Number", "", 48, 2
          SetMFTValueInfo .attrValues(15), "Update Sequence Array.", "Length = S*2-2. See offset 6.", 50 + 2, size * 2 - 2
       End If
   End With
   ReDim info.recordAttributes(0)
   If attrOffset <= 0 Then
      Erase abRecord
      Exit Function
   End If
'*************attributes****************
    Dim pos As Long, Count As Long, n As Long, lngTemp As Long
    Dim dataRunsOffset As Long
    Dim intTemp As Integer, intTemp1 As Integer
    Dim ah As ATTR_HEADER
    Dim attrStandard As ATTR_STANDARD
    Dim attrFileName As ATTR_FILENAME
    Dim SecurityOffsets(3) As Long
    Dim aclSize As Long
    Dim li As LARGE_INTEGER
    
    pos = attrOffset
    Do
      CopyMemory ah, abRecord(pos), Len(ah)
      If ah.dwType = &HFFFFFFFF Then Exit Do
      If Count > 0 Then ReDim Preserve info.recordAttributes(Count)
      attrSize = ah.dwFullLength
      With info.recordAttributes(Count)
         SetMFTValueDirect .attrName, "Attribute type", GetAttrDescription(ah.dwType), "$" & Hex(ah.dwType) & " (" & TranslateAttrType(ah.dwType) & ")", pos, ah.dwFullLength
         ReDim .attrValues(5)
         SetMFTValueInfo .attrValues(0), "Attribute length", "including header", pos + 4, 4
         s = ah.uchNonResFlag & " (" & IIf(ah.uchNonResFlag = 0, "Resident)", "Non resident)")
         SetMFTValueDirect .attrValues(1), "Recidency flag", "0: resident, 1: in runs", s, pos + 8, 1
         SetMFTValueInfo .attrValues(2), "Length of name", "if used", pos + 9, 1
         SetMFTValueInfo .attrValues(3), "Offset to name or resident data", "", pos + 10, 2
         SetMFTValueInfo .attrValues(4), "Compressed flag", "0x0: normal, 0x1: compressed, 0x4000: encrypted, 0x8000: sparse", pos + 12, 1
         SetMFTValueInfo .attrValues(5), "Identificator", "Attribute ID", pos + 14, 2
         n = UBound(.attrValues)
         If ah.uchNonResFlag Then
            ReDim Preserve .attrValues(n + 7)
            .attrDataRuns.attrType = ah.dwType
            
            CopyMemory li, abRecord(pos + 16), 8
            .attrDataRuns.StartVCN = LI_Decimal_LE(li)
            SetMFTValueDirect .attrValues(n + 1), "Start VCN", "", .attrDataRuns.StartVCN, pos + 16, 8
            
            CopyMemory li, abRecord(pos + 24), 8
            .attrDataRuns.EndVCN = LI_Decimal_LE(li)
            SetMFTValueDirect .attrValues(n + 2), "Last VCN", "", .attrDataRuns.StartVCN, pos + 24, 8
            
            CopyMemory dataRunsOffset, abRecord(pos + 32), 2
            SetMFTValueInfo .attrValues(n + 3), "Offset to data runs", "", pos + 32, 2
            SetMFTValueInfo .attrValues(n + 4), "Compression unit size", "2^x clusters, 0: uncompressed", pos + 34, 2
            
            CopyMemory li, abRecord(pos + 40), 8
            .attrDataRuns.allocatedDataSize = LI_Decimal_LE(li)
            SetMFTValueDirect .attrValues(n + 5), "Allocated data size", "rounded to cluster size", .attrDataRuns.allocatedDataSize, pos + 40, 8
            
            CopyMemory li, abRecord(pos + 48), 8
            .attrDataRuns.realDataSize = LI_Decimal_LE(li)
            SetMFTValueDirect .attrValues(n + 6), "Real data size", "", .attrDataRuns.realDataSize, pos + 48, 8
            
            CopyMemory li, abRecord(pos + 56), 8
            .attrDataRuns.streamSize = LI_Decimal_LE(li)
            SetMFTValueDirect .attrValues(n + 7), "Data stream size", "same as real size", .attrDataRuns.streamSize, pos + 56, 8
            
            n = UBound(.attrValues)
            If ah.uchNameLength > 0 Then
               ReDim Preserve .attrValues(n + 1)
               ReDim abTemp(ah.uchNameLength * 2 - 1)
               CopyMemory abTemp(0), abRecord(pos + ah.wNameOffset), ah.uchNameLength * 2
               s = abTemp
               SetMFTValueDirect .attrValues(n + 1), "Attribute name", "Unicode", s, pos + ah.wNameOffset, ah.uchNameLength * 2
            End If
            n = UBound(.attrValues)
            ReDim Preserve .attrValues(n + 1)
            s = currentVolume.GetDataRunsString(mftNum, ah.dwType) ', dataRunsOffset)
            SetMFTValueDirect .attrValues(n + 1), "Data runs", "", s, pos + dataRunsOffset, Len(s) \ 2
            Call GetDataRuns(pos + dataRunsOffset, .attrDataRuns, mftNum)
         Else
            ReDim Preserve .attrValues(n + 3)
            CopyMemory attrDataSize, abRecord(pos + 16), 4
            CopyMemory attrOffset, abRecord(pos + 20), 2
            If attrDataSize > 0 Then
               ReDim .attrResidentData(attrDataSize - 1)
               CopyMemory .attrResidentData(0), abRecord(pos + attrOffset), attrDataSize
            End If
            SetMFTValueInfo .attrValues(n + 1), "Attribute length", "without header", pos + 16, 4
            SetMFTValueInfo .attrValues(n + 2), "Offset to attribute data", "", pos + 20, 2
            SetMFTValueInfo .attrValues(n + 3), "Indexed flag", "", pos + 22, 1
            If ah.uchNameLength > 0 Then
               ReDim Preserve .attrValues(n + 4)
               ReDim abTemp(ah.uchNameLength * 2 - 1)
               CopyMemory abTemp(0), abRecord(pos + ah.wNameOffset), ah.uchNameLength * 2
               s = abTemp
               SetMFTValueDirect .attrValues(n + 4), "Attribute name", "Unicode", s, pos + ah.wNameOffset, ah.uchNameLength * 2
            End If
         End If
         n = UBound(.attrValues)
         Select Case ah.dwType
                Case 0:   ' UNUSED
                Case &H10 ' STANDARD_INFORMATION. Always resident. 48 to 72 bytes
                     ReDim Preserve .attrValues(n + 8)
                     CopyMemory attrStandard, abRecord(pos + attrOffset), Len(attrStandard)
                     SetMFTValueDirect .attrValues(n + 1), "Date created", "In FileTime format", FileTimeToVBDate(attrStandard.n64Create), pos + attrOffset, 8
                     SetMFTValueDirect .attrValues(n + 2), "Date modified", "In FileTime format", FileTimeToVBDate(attrStandard.n64Modify), pos + attrOffset + 8, 8
                     SetMFTValueDirect .attrValues(n + 3), "Date record modified", "In FileTime format", FileTimeToVBDate(attrStandard.n64Modfil), pos + attrOffset + 16, 8
                     SetMFTValueDirect .attrValues(n + 4), "Date last access", "In FileTime format", FileTimeToVBDate(attrStandard.n64Access), pos + attrOffset + 24, 8
                     SetMFTValueDirect .attrValues(n + 5), "File attributes", "", "0x" & Hex(attrStandard.dwFATAttributes) & " (" & ParseFileAttributes(attrStandard.dwFATAttributes) & ")", pos + attrOffset + 32, 4
                     SetMFTValueDirect .attrValues(n + 6), "Max number of versions", "Maximum allowed versions for file. Zero means that version numbering is disabled.", attrStandard.dwMaxVersions, pos + attrOffset + 36, 4
                     SetMFTValueDirect .attrValues(n + 7), "Version number", "This file's version (if any). Will be zero if Maximum Number of Versions is zero.", attrStandard.dwVersionNum, pos + attrOffset + 40, 4
                     SetMFTValueDirect .attrValues(n + 8), "ClassID", "Class Id from bidirectional Class Id index.", attrStandard.dwClassID, pos + attrOffset + 44, 4
                     If attrDataSize > 48 Then
                        ReDim Preserve .attrValues(n + 9)
                        SetMFTValueDirect .attrValues(n + 9), "OwnerID", "Owner Id of the user owning the file. This Id is a key in the $O and $Q Indexes of the file $Quota. If zero, then quotas are disabled.", attrStandard.dwOwnerID, pos + attrOffset + 48, 4
                     End If
                     If attrDataSize > 52 Then
                        ReDim Preserve .attrValues(n + 10)
                        SetMFTValueDirect .attrValues(n + 10), "SecurityID", "This should not be confused with a Security Identifier. The Security Id is a key in the $SII Index and $SDS Data Stream in the file $Secure.", attrStandard.dwSecurityID, pos + attrOffset + 52, 4
                     End If
                     If attrDataSize > 56 Then
                        ReDim Preserve .attrValues(n + 11)
                        SetMFTValueInfo .attrValues(n + 11), "Quota charged", "The number of bytes this file user from the user's quota. This should be the total data size of all streams. If zero, then quotas are disabled.", pos + attrOffset + 56, 8
                     End If
                     If attrDataSize > 64 Then
                        ReDim Preserve .attrValues(n + 12)
                        SetMFTValueInfo .attrValues(n + 12), "Update sequence number", "Last Update Sequence Number of the file. This is a direct index into the file $UsnJrnl. If zero, the USN Journal is disabled.", pos + attrOffset + 64, 8
                     End If
                Case &H20 'ATTR_LIST
                Case &H30 'FILE_NAME
                     If ah.uchNonResFlag = 0 Then
                        ReDim Preserve .attrValues(n + 12)
                        CopyMemory attrFileName, abRecord(pos + attrOffset), attrDataSize
                        s = Left(StrConv(attrFileName.wFilename, vbFromUnicode), attrFileName.chFileNameLength)
                        SetMFTValueDirect .attrValues(n + 1), "Parent MFT record number", "", attrFileName.dwMftParentDir.dwNumLow, pos + attrOffset, 6
                        SetMFTValueDirect .attrValues(n + 2), "Parent MFT record sequence", "", attrFileName.dwMftParentDir.wSequence, pos + attrOffset + 6, 2
                        SetMFTValueDirect .attrValues(n + 3), "Date created", "In FileTime format", FileTimeToVBDate(attrFileName.n64Create), pos + attrOffset + 8, 8
                        SetMFTValueDirect .attrValues(n + 4), "Date modified", "In FileTime format", FileTimeToVBDate(attrFileName.n64Modify), pos + attrOffset + 16, 8
                        SetMFTValueDirect .attrValues(n + 5), "Date record modified", "In FileTime format", FileTimeToVBDate(attrFileName.n64Modfil), pos + attrOffset + 24, 8
                        SetMFTValueDirect .attrValues(n + 6), "Date last access", "In FileTime format", FileTimeToVBDate(attrFileName.n64Access), pos + attrOffset + 32, 8
                        SetMFTValueDirect .attrValues(n + 7), "Allocated size", "", attrFileName.n64Allocated.LowPart, pos + attrOffset + 40, 8
                        SetMFTValueDirect .attrValues(n + 7), "Real size", "", attrFileName.n64RealSize.LowPart, pos + attrOffset + 48, 8
                        SetMFTValueDirect .attrValues(n + 8), "File attributes", "", "0x" & Hex(attrFileName.dwFlags) & " (" & ParseFileAttributes(attrFileName.dwFlags) & ")", pos + attrOffset + 56, 4
                        SetMFTValueDirect .attrValues(n + 9), "Security ID", "", attrFileName.dwSecurityID, pos + attrOffset + 60, 4
                        SetMFTValueDirect .attrValues(n + 10), "File name length", "In chars", attrFileName.chFileNameLength, pos + attrOffset + 64, 1
                        SetMFTValueDirect .attrValues(n + 11), "File name type", "1: Win32, 2: DOS, 3: Win32 & DOS", attrFileName.chFileNameType, pos + attrOffset + 65, 1
                        SetMFTValueDirect .attrValues(n + 12), "File name", "Unicode", s, pos + attrOffset + 66, Len(s) * 2
                     End If
                Case &H40 ' OBJECT_ID. Always resident.
                     If ah.uchNonResFlag = 0 Then
                        ReDim abTemp(15)
                        If attrDataSize > 0 Then
                           CopyMemory abTemp(0), abRecord(pos + attrOffset), 16
                           s = GetGUIDString(abTemp)
                           ReDim Preserve .attrValues(n + 1)
                           SetMFTValueDirect .attrValues(n + 1), "Object ID", "Unique Id assigned to file", s, pos + attrOffset, attrDataSize
                        End If
                        If attrDataSize > 16 Then
                           CopyMemory abTemp(0), abRecord(pos + attrOffset + 16), 16
                           s = GetGUIDString(abTemp)
                           ReDim Preserve .attrValues(n + 2)
                           SetMFTValueDirect .attrValues(n + 2), "Birth Volume GUID", "Volume where file was created", s, pos + attrOffset, attrDataSize
                        End If
                        If attrDataSize > 32 Then
                           CopyMemory abTemp(0), abRecord(pos + attrOffset + 32), 16
                           s = GetGUIDString(abTemp)
                           ReDim Preserve .attrValues(n + 3)
                           SetMFTValueDirect .attrValues(n + 3), "Birth Object GUID", "Original Object Id of file", s, pos + attrOffset, attrDataSize
                        End If
                        If attrDataSize > 48 Then
                           CopyMemory abTemp(0), abRecord(pos + attrOffset + 48), 16
                           s = GetGUIDString(abTemp)
                           ReDim Preserve .attrValues(n + 4)
                           SetMFTValueDirect .attrValues(n + 4), "Domain GUID", "Domain in which object was created", s, pos + attrOffset, attrDataSize
                        End If
                      End If
                Case &H50 ' SECURITY_DESCRIPTOR
                      If ah.uchNonResFlag = 0 Then
                         ReDim Preserve .attrValues(n + 10)
                         SetMFTValueInfo .attrValues(n + 1), "Revision", "", pos + attrOffset, 1
                         CopyMemory intTemp, abRecord(pos + attrOffset + 2), 2
                         s = "0x" & PadLeft(Hex(intTemp), 4) & " (" & ParseSecurityFlags(intTemp) & ")"
                         SetMFTValueDirect .attrValues(n + 2), "Security flags", "", s, pos + attrOffset + 2, 2
                         CopyMemory SecurityOffsets(0), abRecord(pos + attrOffset + 4), 16
                         SetMFTValueDirect .attrValues(n + 3), "Offset to User SID", "Excluding attribute header.", SecurityOffsets(0) & " (0x" & Hex(SecurityOffsets(0)) & ")", pos + attrOffset + 4, 4
                         SetMFTValueDirect .attrValues(n + 4), "Offset to Group SID", "Excluding attribute header.", SecurityOffsets(1) & " (0x" & Hex(SecurityOffsets(1)) & ")", pos + attrOffset + 8, 4
                         SetMFTValueDirect .attrValues(n + 5), "Offset to SACL", "Excluding attribute header. 0 if SACL not presents.", SecurityOffsets(2) & " (0x" & Hex(SecurityOffsets(2)) & ")", pos + attrOffset + 12, 4
                         SetMFTValueDirect .attrValues(n + 6), "Offset to DACL", "Excluding attribute header. 0 if DACL not presents.", SecurityOffsets(3) & " (0x" & Hex(SecurityOffsets(3)) & ")", pos + attrOffset + 16, 4
                         If intTemp And &H1 Then 'Owner defaulted
                            s = "Owner defaulted"
                         Else
                            s = GetSIDString(pos + attrOffset + SecurityOffsets(0))
                         End If
                         SetMFTValueDirect .attrValues(n + 7), "User SID", GetSIDDescription(s), s, pos + attrOffset + SecurityOffsets(0), (UBound(Split(s, "-")) - 2) * 4 + 8
                         If intTemp And &H2 Then 'Group defaulted
                            s = "Group defaulted"
                         Else
                            s = GetSIDString(pos + attrOffset + SecurityOffsets(1))
                         End If
                         SetMFTValueDirect .attrValues(n + 8), "Group SID", GetSIDDescription(s), s, pos + attrOffset + SecurityOffsets(1), (UBound(Split(s, "-")) - 2) * 4 + 8
                         If (intTemp And &H10) Then 'SACL presents
                            If SecurityOffsets(2) = 0 Then
                               s = "NULL ACL"
                               s1 = ""
                               aclSize = 0
                            Else
                               CopyMemory aclSize, abRecord(pos + attrOffset + SecurityOffsets(2) + 2), 2
                               s = GetACLString(pos + attrOffset + SecurityOffsets(2))
                               s1 = GetACLDescription(pos + attrOffset + SecurityOffsets(2))
                            End If
                         Else
                            aclSize = 0
                            s = "missing"
                            s1 = ""
                         End If
                         SetMFTValueDirect .attrValues(n + 9), "System ACL", s1, s, pos + attrOffset + SecurityOffsets(2), aclSize
                         If (intTemp And &H4) Then 'DACL presents
                            If SecurityOffsets(3) = 0 Then
                               s = "NULL ACL"
                               s1 = ""
                               aclSize = 0
                            Else
                               CopyMemory aclSize, abRecord(pos + attrOffset + SecurityOffsets(3) + 2), 2
                               s = GetACLString(pos + attrOffset + SecurityOffsets(3))
                               s1 = GetACLDescription(pos + attrOffset + SecurityOffsets(3))
                            End If
                         Else
                            aclSize = 0
                            s = "missing"
                            s1 = ""
                         End If
                         SetMFTValueDirect .attrValues(n + 10), "Discretionary ACL", s1, s, pos + attrOffset + SecurityOffsets(3), aclSize
                      End If
                Case &H60 ' VOLUME_NAME. Always resident.
                        ReDim Preserve .attrValues(n + 1)
                        If attrDataSize > 0 Then
                           ReDim abTemp(attrDataSize - 1)
                           CopyMemory abTemp(0), abRecord(pos + attrOffset), attrDataSize
                           s = abTemp
                        Else
                           s = ""
                        End If
                        SetMFTValueDirect .attrValues(n + 1), "Volume name", "", s, pos + attrOffset, attrDataSize
                Case &H70 ' VOLUME_INFORMATION. Always resident.
                        ReDim Preserve .attrValues(n + 3)
                        SetMFTValueInfo .attrValues(n + 1), "Major NTFS version", "", pos + attrOffset + 8, 1
                        SetMFTValueInfo .attrValues(n + 2), "Minor NTFS version", "", pos + attrOffset + 9, 1
                        SetMFTValueInfo .attrValues(n + 3), "Volume flags", "", pos + attrOffset + 10, 2
                Case &H80 '//DATA
                     If ah.uchNonResFlag = 0 Then
                        If attrDataSize > 0 Then
                           ReDim Preserve .attrValues(n + 1)
                           ReDim abTemp(attrDataSize - 1)
                           CopyMemory abTemp(0), abRecord(pos + attrOffset), attrDataSize
                           s = StrConv(abTemp, vbUnicode)
                           SetMFTValueDirect .attrValues(n + 1), "Resident data", "", s, pos + attrOffset, attrDataSize
                        End If
                     End If
                Case &H90  ' INDEX_ROOT. Always resident.
                        ReDim Preserve .attrValues(n + 13)
                        CopyMemory lngTemp, abRecord(pos + attrOffset), 4
                        s = "$" & Hex(lngTemp) & " (" & TranslateAttrType(lngTemp) & ")"
                        SetMFTValueDirect .attrValues(n + 1), "Indexed attributes type", "0x30 (FILENAME) for directories, 0x00 otherwise", s, pos + attrOffset, 4
                        CopyMemory lngTemp, abRecord(pos + attrOffset + 4), 4
                        SetMFTValueDirect .attrValues(n + 2), "Collation rule", "Collation rule used to sort the index entries. If type is $FILENAME, this must be COLLATION_FILENAME", lngTemp & " (0x" & Hex(lngTemp) & ") - " & GetCollationRule(lngTemp), pos + attrOffset + 4, 4
                        SetMFTValueInfo .attrValues(n + 3), "Bytes per index block", "Byte size of each index block (in the index allocation attribute).", pos + attrOffset + 8, 4
                        SetMFTValueInfo .attrValues(n + 4), "Clusters per index block", "Cluster size of each index block (in the index allocation attribute), when an index block is >= than a cluster, otherwise this will be the log of the size. Has to be a power of 2.", pos + attrOffset + 12, 1
                        SetMFTValueInfo .attrValues(n + 5), "First entry offset", "Offset to first Index Entry", pos + attrOffset + 16, 4
                        SetMFTValueInfo .attrValues(n + 6), "Total entries size", "Total size of the Index Entries", pos + attrOffset + 20, 4
                        SetMFTValueInfo .attrValues(n + 7), "Allocated size", "Allocated size of the Index Entries", pos + attrOffset + 24, 4
                        SetMFTValueInfo .attrValues(n + 8), "Flags", "0x00: Small Index (fits in Index Root), 0x01: Large index (Index Allocation needed)", pos + attrOffset + 28, 1
                        SetMFTValueInfo .attrValues(n + 9), "MFT reference", "", pos + attrOffset + 32, 6
                        SetMFTValueInfo .attrValues(n + 10), "MFT sequence", "", pos + attrOffset + 38, 2
                        SetMFTValueInfo .attrValues(n + 11), "Index record length", "", pos + attrOffset + 40, 2
                        SetMFTValueInfo .attrValues(n + 12), "Key length", "Byte size of the key value, which is in the index entry (if any)", pos + attrOffset + 42, 2
                        CopyMemory intTemp, abRecord(pos + attrOffset + 44), 2
                        SetMFTValueDirect .attrValues(n + 13), "Index entry flag", "0x01: Entry have subnodes, 0x02: Last entry in a list (doesn't contain any info except of relative VCN).", intTemp, pos + attrOffset + 44, 2
                        If Not (intTemp And 2) Then
                           n = UBound(.attrValues)
                           ReDim Preserve .attrValues(n + 12)
                           CopyMemory attrFileName, abRecord(pos + attrOffset + 48), Len(attrFileName)
                           s = Left(StrConv(attrFileName.wFilename, vbFromUnicode), attrFileName.chFileNameLength)
                           SetMFTValueInfo .attrValues(n + 1), "Parent MFT record number", "", pos + attrOffset + 48, 6
                           CopyMemory intTemp1, abRecord(pos + attrOffset + 54), 2
                           SetMFTValueDirect .attrValues(n + 2), "Parent MFT record sequence", "", intTemp1, pos + attrOffset + 54, 2
                           SetMFTValueDirect .attrValues(n + 3), "Date created", "In FileTime format", FileTimeToVBDate(attrFileName.n64Create), pos + attrOffset + 56, 8
                           SetMFTValueDirect .attrValues(n + 4), "Date modified", "In FileTime format", FileTimeToVBDate(attrFileName.n64Modify), pos + attrOffset + 64, 8
                           SetMFTValueDirect .attrValues(n + 5), "Date record modified", "In FileTime format", FileTimeToVBDate(attrFileName.n64Modfil), pos + attrOffset + 72, 8
                           SetMFTValueDirect .attrValues(n + 6), "Date last access", "In FileTime format", FileTimeToVBDate(attrFileName.n64Access), pos + attrOffset + 80, 8
                           SetMFTValueDirect .attrValues(n + 7), "Allocated size", "", attrFileName.n64Allocated.LowPart, pos + attrOffset + 88, 8
                           SetMFTValueDirect .attrValues(n + 7), "Real size", "", attrFileName.n64RealSize.LowPart, pos + attrOffset + 96, 8
                           SetMFTValueDirect .attrValues(n + 8), "File attributes", "", "0x" & Hex(attrFileName.dwFlags) & " (" & ParseFileAttributes(attrFileName.dwFlags) & ")", pos + attrOffset + 104, 4
                           SetMFTValueDirect .attrValues(n + 9), "Security ID", "", attrFileName.dwSecurityID, pos + attrOffset + 108, 4
                           SetMFTValueDirect .attrValues(n + 10), "File name length", "In chars", attrFileName.chFileNameLength, pos + attrOffset + 112, 1
                           SetMFTValueDirect .attrValues(n + 11), "File name type", "1: Win32, 2: DOS, 3: Win32 & DOS", attrFileName.chFileNameType, pos + attrOffset + 113, 1
                           SetMFTValueDirect .attrValues(n + 12), "File name", "Unicode", s, pos + attrOffset + 114, Len(s) * 2
                        End If
                        If (intTemp And 1) Then
                            n = UBound(.attrValues)
                            ReDim Preserve .attrValues(n + 1)
                            SetMFTValueInfo .attrValues(n + 1), "Relative VCN", "Virtual cluster number of the index block that holds the entries immediately preceding the current entry (the vcn references the corresponding cluster in the data of the non-resident index allocation attribute)", pos + attrSize - 8, 8
                        End If
                Case &HA0  '//INDEX_ALLOCATION
                Case &HB0  '//BITMAP
                     If ah.uchNonResFlag = 0 Then
                        ReDim Preserve .attrValues(n + 1)
                        ReDim abTemp(attrDataSize - 1)
                        CopyMemory abTemp(0), abRecord(pos + attrOffset), attrDataSize
                        s = "0x"
                        For i = 0 To attrDataSize - 1
                            s = s & Right("0" & Hex(abTemp(i)), 2)
                        Next i
                        SetMFTValueDirect .attrValues(n + 1), "Bitmap", "", s, pos + attrOffset, attrDataSize
                     End If
                Case &HC0  ' REPARSE_POINT
                Case &HD0  ' EA_INFORMATION
                Case &HE0  ' EA_DATA
                Case &HF0  ' PROPERTY_SET
                Case &H100 ' LOGGED_UTILITY_STREAM
                Case &H1000 ' FIRST_USER_DEFINED_ATTRIBUTE
                Case &HFFFF&, &HFFFFFFFF ' END
                   Exit Do
         End Select
      End With
      pos = pos + ah.dwFullLength
      If pos > currentVolume.info.BytesPerMFTRecord - Len(ah) Then Exit Do
      If ah.dwFullLength = 0 Then Exit Do
      Count = Count + 1
   Loop
   Erase abRecord
   Erase abTemp
End Function

Private Sub SetMFTValueInfo(ByRef mvi As MFT_VALUE_INFO, ByVal cap As String, ByVal descr As String, ByVal offset As Long, ByVal size As Long)
    With mvi
       .mviCaption = cap
       .mviDescription = descr
       .mviOffset = offset
       .mviSize = size
    End With
    If size < 0 Then Exit Sub
    Dim abTemp() As Byte
    Dim bTemp As Byte
    Dim intTemp As Integer
    Dim lngTemp As Long
    Dim curTemp As Currency
    Dim i As Long
    Dim s As String
    Select Case size
        Case 0
           mvi.mviValue = ""
        Case 1
           CopyMemory bTemp, abRecord(offset), size
           mvi.mviValue = bTemp & " (0x" & PadLeft(Hex(bTemp), 2) & ")"
        Case 2
           CopyMemory intTemp, abRecord(offset), size
           mvi.mviValue = intTemp & " (0x" & PadLeft(Hex(intTemp), 4) & ")"
        Case 4
           CopyMemory lngTemp, abRecord(offset), size
           mvi.mviValue = lngTemp & " (0x" & PadLeft(Hex(lngTemp), 8) & ")"
        Case 6
           CopyMemory curTemp, abRecord(offset), size
           mvi.mviValue = curTemp * 10000
        Case 8
           CopyMemory curTemp, abRecord(offset), size
           mvi.mviValue = CDec(curTemp) * 10000
        Case Else
           ReDim abTemp(size - 1)
           CopyMemory abTemp(0), abRecord(offset), size
           For i = 0 To size - 1
               s = s & Right("0" & Hex(abTemp(i)), 2)
           Next i
           mvi.mviValue = "0x" & s
    End Select
    Erase abTemp
End Sub

Private Sub SetMFTValueDirect(ByRef mvi As MFT_VALUE_INFO, ByVal cap As String, ByVal descr As String, ByVal value As String, ByVal offset As Long, ByVal size As Long)
    With mvi
       .mviCaption = cap
       .mviDescription = descr
       .mviOffset = offset
       .mviSize = size
       .mviValue = value
    End With
End Sub

Public Function GetRecordAttrData(ByVal recNum As Long, ByVal attrType As Long, ByRef adr As ATTR_DATA_RUNS, ByRef abData() As Byte) As Long
    Dim ret As Long, attrOffset As Long, drOffset As Integer
    Dim size As Long, offset As Long
    Dim ah As ATTR_HEADER
    Dim li As LARGE_INTEGER
'******** Read MFT record*******************
    ret = currentVolume.ReadMFTRecord(recNum, abRecord)
    If ret Then
       GetRecordAttrData = ret
       Erase abRecord
       Exit Function
    End If
'******** fixup*******************
    ret = currentVolume.FixUpMFTRecord(abRecord)
    If ret Then
       GetRecordAttrData = ret
       Erase abRecord
       Exit Function
    End If
    
    attrOffset = currentVolume.FindRecordAttribute(recNum, attrType)
    If attrOffset <= 0 Then
       GetRecordAttrData = 87
       Exit Function
    End If
    CopyMemory ah, abRecord(attrOffset), Len(ah)
    If ah.uchNonResFlag Then
       adr.attrType = ah.dwType
       CopyMemory li, abRecord(attrOffset + 16), 8
       adr.StartVCN = LI_Decimal_LE(li)
       CopyMemory li, abRecord(attrOffset + 24), 8
       adr.EndVCN = LI_Decimal_LE(li)
       CopyMemory drOffset, abRecord(attrOffset + 32), 2
       ret = GetDataRuns(attrOffset + drOffset, adr, recNum)
       CopyMemory li, abRecord(attrOffset + 40), 8
       adr.allocatedDataSize = LI_Decimal_LE(li)
       CopyMemory li, abRecord(attrOffset + 48), 8
       adr.realDataSize = LI_Decimal_LE(li)
       CopyMemory li, abRecord(attrOffset + 56), 8
       adr.streamSize = LI_Decimal_LE(li)
    Else
       CopyMemory size, abRecord(attrOffset + 16), 4
       CopyMemory offset, abRecord(attrOffset + 20), 2
       ReDim abData(size - 1)
       CopyMemory abData(0), abRecord(attrOffset + offset), size
    End If
End Function

Private Function GetDataRuns(ByVal drOffset As Long, _
                             ByRef adr As ATTR_DATA_RUNS, _
                             Optional ByVal mftRecNum As Long) As Long
   Dim drSize As Variant
   Dim LCN As Variant, vcn As Variant, nClusters As Variant
   Dim sizeDataLength As Long, vcnDataLength As Long
   Dim liClusters As LARGE_INTEGER, liVCN As LARGE_INTEGER
   Dim pos As Long, i As Long, j As Long
   Dim b As Byte
   Dim abVCN() As Byte
   Dim bSigned As Boolean, bSparse As Boolean
   
   pos = drOffset
   ReDim adr.dataRuns(0)
   Do
     b = abRecord(pos)
     If b = 0 Then Exit Do
     If b = &HFF Then Exit Do
     sizeDataLength = b And &HF         'First nibble  - bytes allocated for data run lenth value
     vcnDataLength = (b And &HF0) / 16  'Second nibble - bytes allocated for VCN value
     If sizeDataLength > 8 Then Exit Do 'Exceed 8 bit integer (dr corrupted?)
     If vcnDataLength > 8 Then vcnDataLength = 0 'Exceed 8 bit integer (dr corrupted?).
     bSparse = (vcnDataLength = 0)
     ZeroMemory liClusters, 8
     ZeroMemory liVCN, 8
     If i > 0 Then
        ReDim Preserve adr.dataRuns(i)
        ReDim abVCN(7)
        If Not bSparse Then
           CopyMemory abVCN(0), abRecord(pos + sizeDataLength + 1), vcnDataLength
          'vcn are signed variables, so check for negative bit
           If abVCN(vcnDataLength - 1) And &H80 Then 'negative
              For j = 7 To vcnDataLength Step -1
                  abVCN(j) = &HFF
              Next j
              bSigned = True
           Else
              bSigned = False
           End If
        End If
        CopyMemory liVCN, abVCN(0), 8
        vcn = LI_Decimal_LE(liVCN, bSigned)
     Else
        CopyMemory liVCN, abRecord(pos + sizeDataLength + 1), vcnDataLength
        vcn = 0
        'vcn = adr.StartVCN
        LCN = LI_Decimal_LE(liVCN)
     End If
     If sizeDataLength > 0 Then
        CopyMemory liClusters, abRecord(pos + 1), sizeDataLength
     End If
     LCN = LCN + vcn
     'Skip invalid LCN's (accept zero LCN for MFT record #7 - $BOOT) only.
     If (LCN <= currentVolume.info.TotalClusters) And ((LCN > 0) Or mftRecNum = 7) Then
        nClusters = LI_Decimal_LE(liClusters)
        If nClusters < 0 Then nClusters = 0
        If nClusters > currentVolume.info.TotalClusters Then nClusters = 0
        adr.dataRuns(i).nClusters = nClusters
        adr.dataRuns(i).LCN = LCN
        adr.dataRuns(i).vcn = vcn
        adr.dataRuns(i).bSparse = bSparse
     End If
     i = i + 1
     pos = pos + sizeDataLength + vcnDataLength + 1
   Loop
   adr.drCount = i
   Erase abVCN
End Function

Private Function GetSIDString(ByVal iOffset As Long) As String
   Dim s As String
   s = "S-"
   s = s & abRecord(iOffset) & "-" & abRecord(iOffset + 7)
   Dim nCount As Integer, i As Integer
   Dim lngTemp As Long
   nCount = abRecord(iOffset + 1)
   For i = 0 To nCount - 1
       CopyMemory lngTemp, abRecord(iOffset + 8 + i * 4), 4
       s = s & "-" & lngTemp
   Next i
   GetSIDString = s
End Function

Private Function GetACLString(ByVal iOffset As Long) As String
   Dim s As String
   Dim intTemp As Integer
   s = "Revision: " & abRecord(iOffset)
   CopyMemory intTemp, abRecord(iOffset + 4), 2
   s = s & "; ACE's count: " & intTemp
   GetACLString = s
End Function

Private Function GetACLDescription(ByVal iOffset As Long) As String
   Dim s As String
   Dim aclSize As Integer, acesCount As Integer, aceSize As Integer
   Dim i As Integer, pos As Integer
   Dim accessMask As Long
   CopyMemory aclSize, abRecord(iOffset + 2), 2
   CopyMemory acesCount, abRecord(iOffset + 4), 2
   s = "<br><i> Revision</i>: " & abRecord(iOffset)
   If acesCount > 0 Then
      s = s & "<br><i>ACE(s)</i>:<ul style=""margin-top:0;margin-left:0.5em;padding-left:0.5em;margin-bottom:-1em"">"
      pos = 8
      For i = 0 To acesCount - 1
         CopyMemory aceSize, abRecord(iOffset + pos + 2), 2
         CopyMemory accessMask, abRecord(iOffset + pos + 4), 4
         s = s & "<b><li>ACE #" & i & "</b><br><i>Type</i>: 0x" & Hex(abRecord(iOffset + pos)) & " (" & GetACEType(abRecord(iOffset + pos)) & ")"
         s = s & "<br><i>Flags</i>: 0x" & Hex(abRecord(iOffset + pos + 1)) & " (" & ParseACEFlags(abRecord(iOffset + pos + 1)) & ")"
         s = s & "<br><i>Access mask</i>: 0x" & Hex(accessMask) ' & " (" & ParseACEFlags(abRecord(iOffset + pos + 1)) & ")"
         s = s & "<br><i>SID</i>: " & GetSIDString(iOffset + pos + 8)
         s = s & "</li>"
         pos = pos + aceSize
      Next i
      s = s & "</ul>"
   End If
   GetACLDescription = s
End Function
